# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Optional,
    Tuple,
    Union,
    cast,
    overload,
)

import numpy as np
from typing_extensions import Literal

from ..types import Axes, AxisAxes, Shape, ShapeOrScalar
from .base import BaseTensor, unwrap1, unwrap_
from .tensor import Tensor, TensorOrScalar, TensorType

if TYPE_CHECKING:
    from .extensions import NormsMethods  # noqa: F401


def assert_bool(x: Any) -> None:
    if not isinstance(x, Tensor):
        return
    if x.dtype != np.dtype("bool"):
        raise ValueError(f"requires dtype bool, got {x.dtype}, consider t.bool().all()")


class NumPyTensor(BaseTensor):
    __slots__ = ()

    # more specific types for the extensions
    norms: "NormsMethods[NumPyTensor]"

    def __init__(self, raw: "np.ndarray"):  # type: ignore
        super().__init__(raw)

    @property
    def raw(self) -> "np.ndarray":  # type: ignore
        return super().raw

    def numpy(self: TensorType) -> Any:
        a = self.raw.view()
        if a.flags.writeable:
            # without the check, we would attempt to set it on array
            # scalars, and that would fail
            a.flags.writeable = False
        return a

    def item(self) -> Union[int, float, bool]:
        return self.raw.item()  # type: ignore

    @property
    def shape(self: TensorType) -> Shape:
        return cast(Tuple, self.raw.shape)

    def reshape(self: TensorType, shape: Union[Shape, int]) -> TensorType:
        if isinstance(shape, int):
            shape = (shape,)
        return type(self)(self.raw.reshape(shape))

    def astype(self: TensorType, dtype: Any) -> TensorType:
        return type(self)(self.raw.astype(dtype))

    def clip(self: TensorType, min_: float, max_: float) -> TensorType:
        return type(self)(np.clip(self.raw, min_, max_))

    def square(self: TensorType) -> TensorType:
        return type(self)(np.square(self.raw))

    def arctanh(self: TensorType) -> TensorType:
        return type(self)(np.arctanh(self.raw))

    def sum(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        return type(self)(self.raw.sum(axis=axis, keepdims=keepdims))

    def prod(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        return type(self)(self.raw.prod(axis=axis, keepdims=keepdims))

    def mean(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        if self.raw.dtype not in [np.float16, np.float32, np.float64]:
            raise ValueError(
                f"Can only calculate the mean of floating types. Got {self.raw.dtype} instead."
            )
        return type(self)(self.raw.mean(axis=axis, keepdims=keepdims))

    def min(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        return type(self)(self.raw.min(axis=axis, keepdims=keepdims))

    def max(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        return type(self)(self.raw.max(axis=axis, keepdims=keepdims))

    def minimum(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(np.minimum(self.raw, unwrap1(other)))

    def maximum(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(np.maximum(self.raw, unwrap1(other)))

    def argmin(self: TensorType, axis: Optional[int] = None) -> TensorType:
        return type(self)(self.raw.argmin(axis=axis))

    def argmax(self: TensorType, axis: Optional[int] = None) -> TensorType:
        return type(self)(self.raw.argmax(axis=axis))

    def argsort(self: TensorType, axis: int = -1) -> TensorType:
        return type(self)(self.raw.argsort(axis=axis))

    def sort(self: TensorType, axis: int = -1) -> TensorType:
        return type(self)(np.sort(self.raw, axis=axis))

    def topk(
        self: TensorType, k: int, sorted: bool = True
    ) -> Tuple[TensorType, TensorType]:
        idx = np.take(np.argpartition(self.raw, k - 1), np.arange(-k, 0), axis=-1)
        val = np.take_along_axis(self.raw, idx, axis=-1)
        if sorted:
            perm = np.flip(np.argsort(val, axis=-1), axis=-1)
            idx = np.take_along_axis(idx, perm, axis=-1)
            val = np.take_along_axis(self.raw, idx, axis=-1)
        return type(self)(val), type(self)(idx)

    def uniform(
        self: TensorType, shape: ShapeOrScalar, low: float = 0.0, high: float = 1.0
    ) -> TensorType:
        return type(self)(np.random.uniform(low, high, size=shape))

    def normal(
        self: TensorType, shape: ShapeOrScalar, mean: float = 0.0, stddev: float = 1.0
    ) -> TensorType:
        return type(self)(np.random.normal(mean, stddev, size=shape))

    def ones(self: TensorType, shape: ShapeOrScalar) -> TensorType:
        return type(self)(np.ones(shape, dtype=self.raw.dtype))

    def zeros(self: TensorType, shape: ShapeOrScalar) -> TensorType:
        return type(self)(np.zeros(shape, dtype=self.raw.dtype))

    def ones_like(self: TensorType) -> TensorType:
        return type(self)(np.ones_like(self.raw))

    def zeros_like(self: TensorType) -> TensorType:
        return type(self)(np.zeros_like(self.raw))

    def full_like(self: TensorType, fill_value: float) -> TensorType:
        from ..astensor import as_raw_tensor

        fill_value_raw = as_raw_tensor(fill_value)
        if isinstance(fill_value_raw, tuple):
            fill_value_raw, _ = fill_value_raw
        return type(self)(np.full_like(self.raw, fill_value_raw))

    def onehot_like(
        self: TensorType, indices: TensorType, *, value: float = 1
    ) -> TensorType:
        if self.ndim != 2:
            raise ValueError("onehot_like only supported for 2D tensors")
        if indices.ndim != 1:
            raise ValueError("onehot_like requires 1D indices")
        if len(indices) != len(self):
            raise ValueError("length of indices must match length of tensor")
        x = np.zeros_like(self.raw)
        rows = np.arange(len(x))
        x[rows, indices.raw] = value
        return type(self)(x)

    def from_numpy(self: TensorType, a: Any) -> TensorType:
        return type(self)(np.asarray(a))

    def _concatenate(
        self: TensorType, tensors: Iterable[TensorType], axis: int = 0
    ) -> TensorType:
        # concatenates only "tensors", but not "self"
        tensors_ = unwrap_(*tensors)
        return type(self)(np.concatenate(tensors_, axis=axis))

    def _stack(
        self: TensorType, tensors: Iterable[TensorType], axis: int = 0
    ) -> TensorType:
        # stacks only "tensors", but not "self"
        tensors_ = unwrap_(*tensors)
        return type(self)(np.stack(tensors_, axis=axis))

    def transpose(self: TensorType, axes: Optional[Axes] = None) -> TensorType:
        if axes is None:
            axes = tuple(range(self.ndim - 1, -1, -1))
        return type(self)(np.transpose(self.raw, axes=axes))

    def all(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        assert_bool(self)
        return type(self)(self.raw.all(axis=axis, keepdims=keepdims))

    def any(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        assert_bool(self)
        return type(self)(self.raw.any(axis=axis, keepdims=keepdims))

    def logical_and(self: TensorType, other: TensorOrScalar) -> TensorType:
        assert_bool(self)
        assert_bool(other)
        return type(self)(np.logical_and(self.raw, unwrap1(other)))

    def logical_or(self: TensorType, other: TensorOrScalar) -> TensorType:
        assert_bool(self)
        assert_bool(other)
        return type(self)(np.logical_or(self.raw, unwrap1(other)))

    def logical_not(self: TensorType) -> TensorType:
        assert_bool(self)
        return type(self)(np.logical_not(self.raw))

    def exp(self: TensorType) -> TensorType:
        return type(self)(np.exp(self.raw))

    def log(self: TensorType) -> TensorType:
        return type(self)(np.log(self.raw))

    def log2(self: TensorType) -> TensorType:
        return type(self)(np.log2(self.raw))

    def log10(self: TensorType) -> TensorType:
        return type(self)(np.log10(self.raw))

    def log1p(self: TensorType) -> TensorType:
        return type(self)(np.log1p(self.raw))

    def tile(self: TensorType, multiples: Axes) -> TensorType:
        multiples = unwrap1(multiples)
        if len(multiples) != self.ndim:
            raise ValueError("multiples requires one entry for each dimension")
        return type(self)(np.tile(self.raw, multiples))

    def softmax(self: TensorType, axis: int = -1) -> TensorType:
        # for numerical reasons we subtract the max logit
        # (mathematically it doesn't matter!)
        # otherwise exp(logits) might become too large or too small
        logits = self.raw
        logits = logits - logits.max(axis=axis, keepdims=True)
        e = np.exp(logits)
        return type(self)(e / e.sum(axis=axis, keepdims=True))

    def log_softmax(self: TensorType, axis: int = -1) -> TensorType:
        # for numerical reasons we subtract the max logit
        # (mathematically it doesn't matter!)
        # otherwise exp(logits) might become too large or too small
        logits = self.raw
        logits = logits - logits.max(axis=axis, keepdims=True)
        log_sum_exp = np.log(np.exp(logits).sum(axis=axis, keepdims=True))
        return type(self)(logits - log_sum_exp)

    def squeeze(self: TensorType, axis: Optional[AxisAxes] = None) -> TensorType:
        return type(self)(self.raw.squeeze(axis=axis))

    def expand_dims(self: TensorType, axis: int) -> TensorType:
        return type(self)(np.expand_dims(self.raw, axis=axis))

    def full(self: TensorType, shape: ShapeOrScalar, value: float) -> TensorType:
        return type(self)(np.full(shape, value, dtype=self.raw.dtype))

    def index_update(
        self: TensorType, indices: Any, values: TensorOrScalar
    ) -> TensorType:
        indices, values = unwrap_(indices, values)
        if isinstance(indices, tuple):
            indices = unwrap_(*indices)
        x = self.raw.copy()
        x[indices] = values
        return type(self)(x)

    def arange(
        self: TensorType,
        start: int,
        stop: Optional[int] = None,
        step: Optional[int] = None,
    ) -> TensorType:
        return type(self)(np.arange(start, stop, step))

    def cumsum(self: TensorType, axis: Optional[int] = None) -> TensorType:
        return type(self)(self.raw.cumsum(axis=axis))

    def flip(self: TensorType, axis: Optional[AxisAxes] = None) -> TensorType:
        return type(self)(np.flip(self.raw, axis=axis))

    def meshgrid(
        self: TensorType, *tensors: TensorType, indexing: str = "xy"
    ) -> Tuple[TensorType, ...]:
        tensors = unwrap_(*tensors)
        outputs = np.meshgrid(self.raw, *tensors, indexing=indexing)
        return tuple(type(self)(out) for out in outputs)

    def pad(
        self: TensorType,
        paddings: Tuple[Tuple[int, int], ...],
        mode: str = "constant",
        value: float = 0,
    ) -> TensorType:
        if len(paddings) != self.ndim:
            raise ValueError("pad requires a tuple for each dimension")
        for p in paddings:
            if len(p) != 2:
                raise ValueError("pad requires a tuple for each dimension")
        if not (mode == "constant" or mode == "reflect"):
            raise ValueError("pad requires mode 'constant' or 'reflect'")
        if mode == "reflect":
            # PyTorch's pad has limited support for 'reflect' padding
            if self.ndim != 3 and self.ndim != 4:
                raise NotImplementedError  # pragma: no cover
            k = self.ndim - 2
            if paddings[:k] != ((0, 0),) * k:
                raise NotImplementedError  # pragma: no cover
        if mode == "constant":
            return type(self)(
                np.pad(self.raw, paddings, mode=mode, constant_values=value)
            )
        else:
            return type(self)(np.pad(self.raw, paddings, mode=mode))

    def isnan(self: TensorType) -> TensorType:
        return type(self)(np.isnan(self.raw))

    def isinf(self: TensorType) -> TensorType:
        return type(self)(np.isinf(self.raw))

    def crossentropy(self: TensorType, labels: TensorType) -> TensorType:
        if self.ndim != 2:
            raise ValueError("crossentropy only supported for 2D logits tensors")
        if self.shape[:1] != labels.shape:
            raise ValueError("labels must be 1D and must match the length of logits")
        # for numerical reasons we subtract the max logit
        # (mathematically it doesn't matter!)
        # otherwise exp(logits) might become too large or too small
        logits = self.raw
        logits = logits - logits.max(axis=1, keepdims=True)
        e = np.exp(logits)
        s = np.sum(e, axis=1)
        ces = np.log(s) - np.take_along_axis(
            logits, labels.raw[:, np.newaxis], axis=1
        ).squeeze(axis=1)
        return type(self)(ces)

    def slogdet(self: TensorType) -> Tuple[TensorType, TensorType]:
        sign, logabsdet = np.linalg.slogdet(self.raw)
        return type(self)(sign), type(self)(logabsdet)

    @overload
    def _value_and_grad_fn(
        self: TensorType, f: Callable[..., TensorType]
    ) -> Callable[..., Tuple[TensorType, TensorType]]:
        ...

    @overload  # noqa: F811 (waiting for pyflakes > 2.1.1)
    def _value_and_grad_fn(
        self: TensorType, f: Callable[..., TensorType], has_aux: Literal[False]
    ) -> Callable[..., Tuple[TensorType, TensorType]]:
        ...

    @overload  # noqa: F811 (waiting for pyflakes > 2.1.1)
    def _value_and_grad_fn(
        self: TensorType,
        f: Callable[..., Tuple[TensorType, Any]],
        has_aux: Literal[True],
    ) -> Callable[..., Tuple[TensorType, Any, TensorType]]:
        ...

    def _value_and_grad_fn(  # noqa: F811 (waiting for pyflakes > 2.1.1)
        self: TensorType, f: Callable, has_aux: bool = False
    ) -> Callable[..., Tuple]:
        # TODO: maybe implement this using https://github.com/HIPS/autograd
        raise NotImplementedError  # pragma: no cover

    def sign(self: TensorType) -> TensorType:
        return type(self)(np.sign(self.raw))

    def sqrt(self: TensorType) -> TensorType:
        return type(self)(np.sqrt(self.raw))

    def tanh(self: TensorType) -> TensorType:
        return type(self)(np.tanh(self.raw))

    def float32(self: TensorType) -> TensorType:
        return self.astype(np.float32)

    def float64(self: TensorType) -> TensorType:
        return self.astype(np.float64)

    def where(self: TensorType, x: TensorOrScalar, y: TensorOrScalar) -> TensorType:
        x, y = unwrap_(x, y)
        return type(self)(np.where(self.raw, x, y))

    def __lt__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__lt__(unwrap1(other)))

    def __le__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__le__(unwrap1(other)))

    def __eq__(self: TensorType, other: TensorOrScalar) -> TensorType:  # type: ignore
        return type(self)(self.raw.__eq__(unwrap1(other)))

    def __ne__(self: TensorType, other: TensorOrScalar) -> TensorType:  # type: ignore
        return type(self)(self.raw.__ne__(unwrap1(other)))

    def __gt__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__gt__(unwrap1(other)))

    def __ge__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__ge__(unwrap1(other)))

    def __getitem__(self: TensorType, index: Any) -> TensorType:
        if isinstance(index, tuple):
            index = tuple(x.raw if isinstance(x, Tensor) else x for x in index)
        elif isinstance(index, Tensor):
            index = index.raw
        return type(self)(self.raw[index])

    def take_along_axis(self: TensorType, index: TensorType, axis: int) -> TensorType:
        if axis % self.ndim != self.ndim - 1:
            raise NotImplementedError(
                "take_along_axis is currently only supported for the last axis"
            )
        return type(self)(np.take_along_axis(self.raw, index.raw, axis=axis))

    def bool(self: TensorType) -> TensorType:
        return self.astype(np.dtype("bool"))



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute exponentially weighted covariance."""
import haiku as hk
import jax.numpy as jnp

from wax.modules.ewma import EWMA


class EWMCov(hk.Module):
    """Compute exponentially weighted covariance.
    To calculate the variance we use the fact that Var(X) = Mean(x^2) - Mean(x)^2 and internally
    we use the exponentially weighted mean of x/x^2 to calculate this.

    References
    ----------
    [^1]: [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf) # noqa
    """

    def __init__(
        self,
        alpha: float = 0.5,
        adjust: bool = True,
        assume_centered: bool = False,
        name: str = None,
    ):
        """

        Args:
            alpha: alpha parameter of the exponential moving average.
            adjust: if true, implement a non-stationary filter with exponential initialization
                scheme. If "linear", implement a non-stationary filter with linear initialization.
            assume_centered: if true, assume that the mean estimator is zero.
            name : name of the module instance.
        """
        super().__init__(name=name)
        self.alpha = alpha
        self.adjust = adjust
        self.assume_centered = assume_centered

    def __call__(self, x, y=None):
        """Compute"""
        if y is None:
            import warnings

            warnings.warn(
                "x and y arguments may be passed directly"
                "instead as a tuple argument. The tuple argument syntax"
                "may be removed in the future."
            )
            x, y = x
        mean_xy = EWMA(self.alpha, self.adjust, initial_value=jnp.nan, name="mean_xy")(
            jnp.outer(x, y)
        )
        if self.assume_centered:
            cov = mean_xy
        else:
            mean_x = EWMA(
                self.alpha, self.adjust, initial_value=jnp.nan, name="mean_x"
            )(x)
            mean_y = EWMA(
                self.alpha, self.adjust, initial_value=jnp.nan, name="mean_y"
            )(y)
            cov = mean_xy - jnp.outer(mean_x, mean_y)
        return cov



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Relative change between the current and a prior element."""
import haiku as hk

from wax.modules import Ffill, Lag


class PctChange(hk.Module):
    """Relative change between the current and a prior element."""

    def __init__(
        self,
        periods: int = 1,
        fill_method: str = "pad",
        limit: int = None,
        name: str = None,
    ):
        super().__init__(name=name)
        self.periods = periods
        self.fill_method = fill_method
        self.limit = limit
        assert periods == 1, "periods > 1 not implemented."

    def __call__(self, x):
        if self.fill_method in ["ffill", "pad"]:
            previous_x = Lag(self.periods)(Ffill()(x))
        else:
            previous_x = Lag(self.periods)()
        pct_change = x / previous_x - 1.0
        return pct_change



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Online optimizer module."""
from typing import Any, Callable, NamedTuple, Union

import haiku as hk
import jax
import jax.numpy as jnp
import optax
from jax.tree_util import tree_map
from optax import GradientTransformation


class ParamsState(NamedTuple):
    trainable_params: Any
    non_trainable_params: Any
    state: Any


class OptimizerOutput(NamedTuple):
    loss: Any
    model_info: Any
    grads: Any
    updated_grads: Any
    trainable_params: Any
    updated_trainable_params: Any
    params: Any
    updated_params: Any
    opt_state: Any


class OnlineOptimizer(hk.Module):
    """Wraps a model with loss and and optimizer to perform one online learning update."""

    def __init__(
        self,
        model: Union[Callable, hk.TransformedWithState],
        opt: GradientTransformation,
        project_params: Callable = None,
        regularize_loss: Callable = None,
        split_params: Callable = None,
        name: str = None,
    ):
        """Initialize module.

        Args:
            model : model to optimize. The model should return a tuple (loss, info)
            opt : optimizer: Optax transformation consisting of a function pair: (initialise, update).
            project_params : function to project parameters. It applies to parameters and optimizer state .
            regularize_loss: function to regularize the model loss. It applies to the parameters.
            split_params: function to split params in trainable and non-trainable params.
                See https://dm-haiku.readthedocs.io/en/latest/notebooks/non_trainable.html
            name : name of the module
        """
        super().__init__(name=name)
        self.model = (
            model
            if isinstance(model, hk.TransformedWithState)
            else hk.transform_with_state(model)
        )
        self.opt = opt
        self.project_params = project_params
        self.regularize_loss = regularize_loss
        self.split_params = (
            split_params
            if split_params is not None
            else lambda params: (params, type(params)())
        )

    def __call__(self, *args, **kwargs):
        """Update learner.

        Args:
            x : features
            y: target
        """

        def init_model_params_and_state(shape, dtype):
            """Set state from trainable params and state of the model."""
            params, state = self.model.init(hk.next_rng_key(), *args, **kwargs)
            trainable_params, non_trainable_params = self.split_params(params)
            trainable_params = hk.data_structures.to_mutable_dict(trainable_params)
            return ParamsState(trainable_params, non_trainable_params, state)

        trainable_params, non_trainable_params, state = hk.get_state(
            "model_params_and_state",
            [],
            init=init_model_params_and_state,
        )

        def init_opt_state(shape, dtype):
            return self.opt.init(trainable_params)

        opt_state = hk.get_state("opt_state", [], init=init_opt_state)

        step = hk.get_state("step", [], init=lambda *_: 0)

        @jax.jit
        def _loss(trainable_params, non_trainable_params, state, *args, **kwargs):
            params = hk.data_structures.merge(trainable_params, non_trainable_params)
            (loss, model_info), state = self.model.apply(
                params, state, None, *args, **kwargs
            )
            if self.regularize_loss:
                loss += self.regularize_loss(trainable_params)
            return loss

        # compute loss and gradients
        l, grads = jax.value_and_grad(_loss)(
            trainable_params, non_trainable_params, state, *args, **kwargs
        )

        # compute prediction and update model state
        params = hk.data_structures.merge(trainable_params, non_trainable_params)
        (loss, model_info), state = self.model.apply(
            params, state, None, *args, **kwargs
        )

        # update optimizer state
        filled_grads = tree_map(jnp.nan_to_num, grads)
        updated_grads, opt_state = self.opt.update(filled_grads, opt_state)

        # update params
        updated_trainable_params = optax.apply_updates(trainable_params, updated_grads)

        if self.project_params:
            updated_trainable_params = self.project_params(
                updated_trainable_params, opt_state
            )

        updated_params = hk.data_structures.merge(
            updated_trainable_params, non_trainable_params
        )
        step += 1
        hk.set_state("step", step)
        hk.set_state(
            "model_params_and_state",
            ParamsState(updated_trainable_params, non_trainable_params, state),
        )
        hk.set_state("opt_state", opt_state)
        return OptimizerOutput(
            loss,
            model_info,
            grads,
            updated_grads,
            trainable_params,
            updated_trainable_params,
            params,
            updated_params,
            opt_state=opt_state,
        )



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implement difference of values on sequential data."""
import haiku as hk

from wax.modules.buffer import Buffer


class Diff(hk.Module):
    """Implement difference of values on sequential data."""

    def __init__(self, periods: int = 1, name: str = None):
        """Initialize module.

        Args:
            periods : number of lags to use to compute the diff.
        """
        super().__init__(name=name)
        self.periods = periods
        assert periods == 1, "periods > 1 not implemented."

    def __call__(self, x):
        """Compute diff.

        Args:
            x: input data
        """
        buffer = Buffer(self.periods + 1)(x)
        diff = buffer[-1] - buffer[0]
        return diff



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
from jax import numpy as jnp

from wax.modules import Ffill
from wax.unroll import unroll


@pytest.mark.parametrize("use_jit", [False, True])
def test_ffill(use_jit):

    xs = jnp.array([90, 91, jnp.nan, 85])
    res = unroll(lambda x: Ffill()(x))(xs)
    assert jnp.allclose(res, jnp.array([90, 91, 91, 85], dtype=jnp.float32))



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING, Tuple, Union

if TYPE_CHECKING:
    # for static analyzers
    import jax  # noqa: F401
    import numpy  # noqa: F401
    import tensorflow  # noqa: F401
    import torch  # noqa: F401

Axes = Tuple[int, ...]
AxisAxes = Union[int, Axes]
Shape = Tuple[int, ...]
ShapeOrScalar = Union[Shape, int]

# tensorflow.Tensor, jax.numpy.ndarray and numpy.ndarray currently evaluate to Any
# we can therefore only provide additional type information for torch.Tensor
NativeTensor = Union[
    "torch.Tensor", "tensorflow.Tensor", "jax.numpy.ndarray", "numpy.ndarray"
]



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax.numpy as jnp

from wax.modules.buffer import Buffer
from wax.modules.fill_nan_inf import FillNanInf


class SNARIMAX(hk.Module):
    """SNARIMAX Adaptive filter.

    It can be used to forecast timeseries with a ARMA dynamic using the online learning
    method as described in [^4].

    The API of this module is similar to the one of SNARIMAX model in the river Python library.

    Parameters
    ----------
    p
        Order of the autoregressive part. This is the number of past target values that will be
        included as features.
    d
        Differencing order.
    q
        Order of the moving average part. This is the number of past error terms that will be
        included as features.
    m
        Season length used for extracting seasonal features. If you believe your data has a
        seasonal pattern, then set this accordingly. For instance, if the data seems to exhibit
        a yearly seasonality, and that your data is spaced by month, then you should set this
        to 12. Note that for this parameter to have any impact you should also set at least one
        of the `p`, `d`, and `q` parameters.
    sp
        Seasonal order of the autoregressive part. This is the number of past target values
        that will be included as features.
    sd
        Seasonal differencing order.
    sq
        Seasonal order of the moving average part. This is the number of past error terms that
        will be included as features.

    regressor
        The online regression model to use. By default, a haiku `Linear` module.

    References
    ----------
    [^1]: [Wikipedia page on ARMA](https://www.wikiwand.com/en/Autoregressive%E2%80%93moving-average_model)
    [^2]: [Wikipedia page on NARX](https://www.wikiwand.com/en/Nonlinear_autoregressive_exogenous_model)
    [^3]: [ARIMA models](https://otexts.com/fpp2/arima.html)
    [^4] [Anava, O., Hazan, E., Mannor, S. and Shamir, O., 2013, June.
    Online learning for time series prediction. In Conference on learning theory (pp. 172-184)]
    (https://arxiv.org/pdf/1302.6927.pdf)

    """

    def __init__(
        self,
        p: int,
        d: int = 0,
        q: int = 0,
        m: int = 1,
        sp: int = 0,
        sd: int = 0,
        sq: int = 0,
        regressor=None,
        name=None,
    ):
        super().__init__(name=name)
        self.p = p
        self.d = d
        self.q = q
        self.m = m
        self.sp = sp
        self.sd = sd
        self.sq = sq
        self.regressor = regressor if regressor is not None else hk.Linear(1)
        if d != 0 or sp != 0 or sd != 0 or sq != 0:
            raise NotImplementedError(
                "Options 'd', 'sp', 'sd', 'sq' are not yet implemented."
            )

    def __call__(self, y, X=None):
        yp = Buffer(self.p + 1, name="y_trues")(y)[1:]

        errp = hk.get_state(
            "errp",
            [],
            init=lambda *_: jnp.full((self.q + 1,) + y.shape, 0.0, y.dtype)[1:],
        )

        X = [X] if X is not None else []
        X += [errp.flatten(), yp.flatten()]
        X = jnp.concatenate(X)
        X = FillNanInf()(X)

        y_pred = self.regressor(X).reshape(y.shape)
        err = y - y_pred
        errp = Buffer(self.q + 1, name="err_lag")(err)[1:]
        hk.set_state("errp", errp)
        return y_pred, {}



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import pandas as pd
import pytest
from jax.config import config

from wax.compile import jit_init_apply
from wax.modules.ewma import EWMA
from wax.unroll import dynamic_unroll_fori_loop, unroll


@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_init_and_first_step_ema(dtype):

    if dtype == "float64":
        config.update("jax_enable_x64", True)
    else:
        config.update("jax_enable_x64", False)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(3,), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(0.1, adjust=True)(x)

    params, state = model.init(next(seq), x)
    ema, state = model.apply(params, state, next(seq), x)
    assert ema.dtype == jnp.dtype(dtype)


def test_run_ema_vs_pandas_not_adjust():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(0.1, adjust=False)(x)

    ema, state = unroll(model, dynamic=False, return_final_state=True)(x)

    pandas_ema = pd.DataFrame(x).ewm(alpha=0.1, adjust=False).mean()

    assert jnp.allclose(ema, pandas_ema.values)


def test_dynamic_unroll_fori_loop():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(0.1, adjust=True)(x)

    ema, state = unroll(model, dynamic=False, return_final_state=True)(x)

    ema2, state2 = dynamic_unroll_fori_loop(model, None, None, next(seq), False, x)

    assert jnp.allclose(ema, ema2)


def test_dynamic_unroll():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(0.1, adjust=True)(x)

    ema, state = unroll(model, dynamic=False, return_final_state=True)(x)

    seq = hk.PRNGSequence(42)
    ema2, state2 = unroll(model, return_final_state=True)(x)

    assert jnp.allclose(ema, ema2)


def test_run_ema_vs_pandas_adjust():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(0.1, adjust=True)(x)

    ema, state = unroll(model, return_final_state=True)(x)

    pandas_ema = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).mean()
    assert jnp.allclose(ema, pandas_ema.values)


def test_run_ema_vs_pandas_adjust_finite():

    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMA(0.1, adjust="linear")(x)

    ema, state = unroll(model, return_final_state=True)(x)
    pandas_ema_adjust = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).mean()
    pandas_ema_not_adjust = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).mean()
    assert not jnp.allclose(ema, pandas_ema_adjust.values)
    assert not jnp.allclose(ema, pandas_ema_not_adjust.values)
    corr = jnp.corrcoef(ema.flatten(), pandas_ema_adjust.values.flatten())[0, 1]
    assert 1.0e-3 < 1 - corr < 1.0e-2


@pytest.mark.parametrize("adjust", [False, True, "linear"])
def test_grad_ewma(adjust):
    from functools import partial

    import jax
    import jax.numpy as jnp

    from wax.unroll import unroll_transform_with_state

    rng = jax.random.PRNGKey(42)
    x = jax.random.normal(rng, (10, 3))
    _, rng = jax.random.split(rng)
    params = {"w": jax.random.normal(rng, (10,))}

    # put some nan values
    x = jax.ops.index_update(x, 0, jnp.nan)

    @partial(unroll_transform_with_state, dynamic=False)
    def fun(x):
        return EWMA(1 / 10, adjust=adjust)(x)

    # print("init")
    params, state = fun.init(rng, x)

    # print("apply")
    res, final_state = fun.apply(params, state, rng, x)
    # print(res)

    # print("gradient")
    @jax.value_and_grad
    def batch(params):
        res, final_state = fun.apply(params, state, rng, x)
        return res.mean()

    score, grad = batch(params)
    assert not jnp.isnan(grad["ewma"]["alpha"])



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Union

import haiku as hk
import jax
from haiku import TransformedWithState


class VMap(hk.Module):
    def __init__(self, fun: Union[Callable, TransformedWithState], name=None):
        super().__init__(name=name)
        self.fun = (
            fun
            if isinstance(fun, TransformedWithState)
            else hk.transform_with_state(fun)
        )

    def __call__(self, *args, **kwargs):
        n_batches = len(jax.tree_leaves((args, kwargs))[0])
        rng = hk.next_rng_key()
        rng = jax.random.split(rng, num=n_batches)
        params, state = hk.get_state(
            "params_state",
            [],
            init=lambda *_: jax.vmap(self.fun.init)(rng, *args, **kwargs),
        )
        res, state = jax.vmap(self.fun.apply)(params, state, rng, *args, **kwargs)
        hk.set_state("params_state", (params, state))

        return res


# Helper functions


def add_batch(fun: Union[Callable, TransformedWithState], take_mean=True):
    """Wrap a function with VMap module.
    It should be used inside a transformed function."""

    def fun_batch(*args, **kwargs):
        res = VMap(fun)(*args, **kwargs)
        if take_mean:
            res = jax.tree_map(lambda x: x.mean(axis=0), res)
        return res

    return fun_batch



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Detect if something has changed."""
import haiku as hk
import jax.numpy as jnp

from wax.stream import DTYPE_INIT_VALUES


class HasChanged(hk.Module):
    """Detect if something has changed."""

    def __call__(self, input):
        prev_input = hk.get_state(
            "prev_value",
            [],
            init=lambda *_: jnp.full_like(
                input, DTYPE_INIT_VALUES[input.dtype.type], input.dtype
            ),
        )
        has_changed = jnp.not_equal(input, prev_input).any()
        hk.set_state("prev_value", input)
        return has_changed



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# + colab={"base_uri": "https://localhost:8080/"} id="2b295407-92c4-4818-bfb9-f445f6967f10" outputId="dc6c8e1b-2875-4287-d83a-4bdc4c9db80a"
# Uncomment to run the notebook in Colab
# # ! pip install -q "wax-ml[complete]@git+https://github.com/eserie/wax-ml.git"
# # ! pip install -q --upgrade jax jaxlib==0.1.70+cuda111 -f https://storage.googleapis.com/jax-releases/jax_releases.html

# + id="ff30291d"
# check available devices
import jax

# + colab={"base_uri": "https://localhost:8080/"} id="a3cdb104" outputId="a6f395c3-6ee3-4fe5-ce39-a02617a129ca"
print("jax backend {}".format(jax.lib.xla_bridge.get_backend().platform))
jax.devices()

# + [markdown] id="1fa1808c"
# # 🎛 The 3-steps workflow 🎛
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/eserie/wax-ml/blob/main/docs/notebooks/04_The_three_steps_workflow.ipynb)

# + [markdown] id="2e1eefec"
# It is already very useful to be able to execute a JAX function on a dataframe in a single work step
# and with a single command line thanks to WAX-ML accessors.
#
# The 1-step WAX-ML's stream API works like that:
# ```python
# <data-container>.stream(...).apply(...)
# ```
#
# But this is not optimal because, under the hood, there are mainly three costly steps:
# - (1) (synchronize | data tracing | encode): make the data "JAX ready"
# - (2) (compile | code tracing | execution): compile and optimize a function for XLA, execute it.
# - (3) (format): convert data back to pandas/xarray/numpy format.
#
# With the `wax.stream` primitives, it is quite easy to explicitly split the 1-step workflow
# into a 3-step workflow.
#
# This will allow the user to have full control over each step and iterate on each one.
#
# It is actually very useful to iterate on step (2), the "calculation step" when
# you are doing research.
# You can then take full advantage of the JAX primitives, especially the `jit` primitive.
#
# Let's illustrate how to reimplement WAX-ML EWMA yourself with the WAX-ML 3-step workflow.

# + [markdown] id="bd2e906e"
# ## Imports

# + id="2bdfbf9a" tags=[]
import numpy as onp
import pandas as pd
import xarray as xr

from wax.accessors import register_wax_accessors
from wax.external.eagerpy import convert_to_tensors
from wax.format import format_dataframe
from wax.modules import EWMA
from wax.stream import tree_access_data
from wax.unroll import unroll

register_wax_accessors()

# + [markdown] id="a0999c69"
# ## Performance on big dataframes

# + [markdown] id="6f8447b4"
# ### Generate data

# + id="768d7802-580d-4c31-9a0e-e6dc4f0589ca" tags=["parameters"]
T = 1.0e5
N = 1000

# + id="03af743d" tags=[]
T, N = map(int, (T, N))
dataframe = pd.DataFrame(
    onp.random.normal(size=(T, N)), index=pd.date_range("1970", periods=T, freq="s")
)

# + [markdown] id="d1fd46f7"
# ### pandas EWMA

# + colab={"base_uri": "https://localhost:8080/"} id="27092faf" outputId="e97a2738-5c2a-4bf9-c9e0-c44040185424" tags=[]
# %%time
df_ewma_pandas = dataframe.ewm(alpha=1.0 / 10.0).mean()

# + [markdown] id="678be283"
# ### WAX-ML EWMA

# + colab={"base_uri": "https://localhost:8080/"} id="11f3705d" outputId="744f0faa-7030-4155-80ae-4a5c745731f8" tags=[]
# %%time
df_ewma_wax = dataframe.wax.ewm(alpha=1.0 / 10.0).mean()

# + [markdown] id="0d94d5cf"
# It's a little faster, but not that much faster...

# + [markdown] id="e51ee290"
# ### WAX-ML EWMA (without format step)

# + [markdown] id="7361def9"
# Let's disable the final formatting step (the output is now in raw JAX format):

# + colab={"base_uri": "https://localhost:8080/"} id="f87f5668" outputId="3f37d97e-1875-4f66-9e6b-9a22b20642a6" tags=[]
# %%time
df_ewma_wax_no_format = dataframe.wax.ewm(alpha=1.0 / 10.0, format_outputs=False).mean()
df_ewma_wax_no_format.block_until_ready()

# + colab={"base_uri": "https://localhost:8080/"} id="88d0cab5-62ad-47f7-9d06-56fc45fa542e" outputId="f8b7c4fb-79a7-4652-82eb-2b649aeb074c"
type(df_ewma_wax_no_format)

# + [markdown] id="9be62475-b2fa-4a95-b293-7e4410ca36ca"
# Let's check the device on which the calculation was performed (if you have GPU available, this should be `GpuDevice` otherwise it will be `CpuDevice`):

# + colab={"base_uri": "https://localhost:8080/"} id="38d3970c-04a5-4deb-93b3-a4f5f899e8f1" outputId="ba287d55-719a-47de-f959-9223effbe7a6"
df_ewma_wax_no_format.device()

# + [markdown] id="784ee16e"
# Now we will see how to break down WAX-ML one-liners `<dataset>.ewm(...).mean()` or `<dataset>.stream(...).apply(...)` into 3 steps:
# - a preparation step where we prepare JAX-ready data and functions.
# - a processing step where we execute the JAX program
# - a post-processing step where we format the results in pandas or xarray format.

# + [markdown] id="c5e7b817"
# ### Generate data (in dataset format)
#
# WAX-ML `Sream` object works on datasets.
# So let's transform the `DataFrame` into a xarray `Dataset`:

# + id="9965444b"
dataset = xr.DataArray(dataframe).to_dataset(name="dataarray")
del dataframe

# + [markdown] id="123965eb"
# ## Step (1) (synchronize | data tracing | encode)
#
# In this step,  WAX-ML do:
# - "data tracing" : prepare the indices for fast access tin the JAX function `access_data`
# - synchronize streams if there is multiple ones.
#   This functionality have options : `freq`, `ffills`
# - encode and convert data from numpy to JAX: use encoders for `datetimes64` and `string_`
#   dtypes. Be aware that by default JAX works in float32
#   (see [JAX's Common Gotchas](https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#double-64bit-precision) to work in float64).
#
# We have a function `Stream.prepare` that implement this Step (1).
# It prepares a function that wraps the input function with the actual data and indices
# in a pair of pure functions (`TransformedWithState` Haiku tuple).

# + colab={"base_uri": "https://localhost:8080/"} id="5f6b72ca" outputId="e6852b1f-65bb-478a-ea6f-1cc6a3ce0898"
# %%time
stream = dataset.wax.stream()

# + [markdown] id="4cc34aad-3e15-4220-9dff-30dcad660307"
# Define our custom function to be applied on a dict of arrays
# having the same structure than the original dataset:


# + id="2bfa2a6c"
def my_ewma_on_dataset(dataset):
    return EWMA(alpha=1.0 / 10.0, adjust=True)(dataset["dataarray"])


# + id="4735903f"
transform_dataset, jxs = stream.prepare(dataset, my_ewma_on_dataset)

# + [markdown] id="18b31346-501a-407b-b586-780117d043f3"
# Let's definite the init parameters and state of the transformation we
# will apply.


# + [markdown] id="903a778c"
# ## Step (2) (compile | code tracing | execution)

# + [markdown] id="425830f4"
# In this step we:
# - prepare a pure function (with
#   [Haiku's transform mechanism](https://dm-haiku.readthedocs.io/en/latest/api.html#haiku-transforms))
#   Define a "transformation" function which:
#     - access to the data
#     - apply another transformation, here: EWMA
#
# - compile it with `jax.jit`
# - perform code tracing and execution (the last line):
#     - Unroll the transformation on "steps" `xs` (a `np.arange` vector).

# + id="6825fdc8-1773-4e04-a41c-a126a1527891"
outputs = unroll(transform_dataset)(jxs)

# + colab={"base_uri": "https://localhost:8080/"} id="3862a486-a5a2-4aa9-967a-59ebc32a18e1" outputId="b54b4e2c-a0ab-4a43-d331-d6b3d90707c6"
outputs.device()

# + [markdown] id="b73f3252"
# Once it has been compiled and "traced" by JAX, the function is much faster to execute:

# + colab={"base_uri": "https://localhost:8080/"} id="a889d294" outputId="e69a75ff-4b00-4a1d-f101-49944e01d9b4" tags=[]
# %%timeit
outputs = unroll(transform_dataset)(jxs)
_ = outputs.block_until_ready()

# + [markdown] id="987e8b63"
# This is 3x faster than pandas implementation!

# + [markdown] id="4c8acec1-4414-4340-9cfc-199e90565d4d"
# ### Manually prepare the data and manage the device

# + [markdown] id="8b76775b-76f7-45ae-8bc7-66a6f945370f"
# In order to manage the device on which the computations take place,
# we need to have even more control over the execution flow.
# Instead of calling `stream.prepare` to build the `transform_dataset` function,
# we can do it ourselves by :
# - using the `stream.trace_dataset` function
# - converting the numpy data in jax ourself
# - puting the data on the device we want.

# + colab={"base_uri": "https://localhost:8080/"} id="76479f76-7b6e-4321-97c3-942f2987bc01" outputId="7f22f580-be0b-4d23-9d4d-7d89c9dd117b"
np_data, np_index, xs = stream.trace_dataset(dataset)
jnp_data, jnp_index, jxs = convert_to_tensors((np_data, np_index, xs), "jax")

# + [markdown] id="a4fa1934-3633-4661-b808-c60e4b8c4600"
# We explicitly set data on CPUs (the is not needed if you only have CPUs):

# + colab={"base_uri": "https://localhost:8080/"} id="bae591f7-e7ec-4b8b-9b80-6f129184e1ae" outputId="36f15d54-170f-4771-f571-ecacbc942026"
from jax.tree_util import tree_leaves, tree_map

cpus = jax.devices("cpu")
jnp_data, jnp_index, jxs = tree_map(
    lambda x: jax.device_put(x, cpus[0]), (jnp_data, jnp_index, jxs)
)
print("data copied to CPU device.")


# + [markdown] id="f851fbe7-096e-4399-ae5c-08739837dfeb"
# We have now "JAX-ready" data for later fast access.

# + [markdown] id="19f3b58e-61f3-481c-a512-cb87bde622a8"
# Let's define the transformation that wrap the actual data and indices in a pair of
# pure functions:

# + id="e7ebbb08-d790-4977-b49d-c9224e299a42"
@jax.jit
@unroll
def transform_dataset(step):
    dataset = tree_access_data(jnp_data, jnp_index, step)
    return EWMA(alpha=1.0 / 10.0, adjust=True)(dataset["dataarray"])


# + [markdown] id="31fa1d05-c3d7-4543-ac82-d398044cdc2e"
# And we can call it as before:

# + colab={"base_uri": "https://localhost:8080/"} id="f851407a-597f-4711-8370-c3e83cb50da7" outputId="3892357f-2d91-40b6-85bc-542e471ed28f"
# %%time
outputs = transform_dataset(jxs)
_ = outputs.block_until_ready()

# + colab={"base_uri": "https://localhost:8080/"} id="f851407a-597f-4711-8370-c3e83cb50da7" outputId="3892357f-2d91-40b6-85bc-542e471ed28f"
# %%time
outputs = transform_dataset(jxs)
_ = outputs.block_until_ready()

# + colab={"base_uri": "https://localhost:8080/"} id="e5b786b7-ecaf-4280-894b-aa8f65a0b78f" outputId="2e3c126e-9c9d-49b4-caa5-67f215944188"
outputs.device()

# + [markdown] id="6fa51498"
# ## Step(3) (format)
# Let's come back to pandas/xarray:

# + colab={"base_uri": "https://localhost:8080/"} id="d1b0732c" outputId="38e747c5-5b1c-40dc-c289-5aeebe503a8b"
# %%time
y = format_dataframe(
    dataset.coords, onp.array(outputs), format_dims=dataset.dataarray.dims
)

# + [markdown] id="e27bd3b8"
# It's quite slow (see WEP3 enhancement proposal).

# + [markdown] id="3cdc9281-6092-4368-a0e8-ed26a5114106"
# ## GPU execution

# + [markdown] id="f0d4651d-9087-4f04-9f07-a4d92cd3ba1f"
# Let's look with execution on GPU

# + colab={"base_uri": "https://localhost:8080/"} id="a30f6024-5e9c-4174-92ea-7207860d829d" outputId="95058ee2-9461-4e7e-f9b1-b1e55087dc9e"
try:
    gpus = jax.devices("gpu")
    jnp_data, jnp_index, jxs = tree_map(
        lambda x: jax.device_put(x, gpus[0]), (jnp_data, jnp_index, jxs)
    )
    print("data copied to GPU device.")
    GPU_AVAILABLE = True
except RuntimeError as err:
    print(err)
    GPU_AVAILABLE = False

# + [markdown] id="27d7cf26-f35b-489e-83b9-c120565d9b17"
# Let's check that our data is on the GPUs:

# + colab={"base_uri": "https://localhost:8080/"} id="d6f00e5d-1b85-483d-9856-35de3a954b13" outputId="9d5d0ec1-4899-4d8e-d9f6-2232321ec42a"
tree_leaves(jnp_data)[0].device()

# + colab={"base_uri": "https://localhost:8080/"} id="6f89fc7c-4824-4858-aee8-6fff7834c70c" outputId="5b5304de-0694-4944-92f1-63eee852ebf6"
tree_leaves(jnp_index)[0].device()

# + colab={"base_uri": "https://localhost:8080/"} id="ba2471ef-51e4-49f2-863c-aabafd401cbf" outputId="e6afa394-3e1b-4b9b-aea6-9dfe0d60ab86"
jxs.device()

# + colab={"base_uri": "https://localhost:8080/"} id="6c35bd5d-0113-455d-bd56-f7c31bf6c736" outputId="fda693e8-723e-436d-84f6-68a49a510be9"
# %%time
if GPU_AVAILABLE:
    outputs = unroll(transform_dataset)(jxs)


# + [markdown] id="274b5615-8785-43e7-a2db-5f867566c913"
# Let's redefine our function `transform_dataset` by explicitly specify to `jax.jit` the `device` option.

# + colab={"base_uri": "https://localhost:8080/"} id="6bb7431e-90f2-4761-a906-69f25fea4a63" outputId="285660f2-8379-4b8f-f89f-9f5a7240c419"
# %%time
from functools import partial

if GPU_AVAILABLE:

    @partial(jax.jit, device=gpus[0])
    @unroll
    def transform_dataset(step):
        dataset = tree_access_data(jnp_data, jnp_index, step)
        return EWMA(alpha=1.0 / 10.0, adjust=True)(dataset["dataarray"])

    outputs = transform_dataset(jxs)

# + colab={"base_uri": "https://localhost:8080/"} id="944dcb8e-211c-4b39-b854-12118fe775ed" outputId="fbeeab1e-e245-4d85-ad5c-09530ec7d331"
outputs.device()

# + colab={"base_uri": "https://localhost:8080/"} id="de99ee71-4e02-4843-9e6e-39d831f9697e" outputId="4262991e-c488-4296-849d-b111e288c203"
# %%timeit
if GPU_AVAILABLE:
    outputs = unroll(transform_dataset)(jxs)
    _ = outputs.block_until_ready()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The online newton optimizer. It extends optax optimizers."""
import jax
import jax.numpy as jnp
from optax._src import base, combine
from optax._src.alias import ScalarOrSchedule, _scale_by_learning_rate


def newton(
    learning_rate: ScalarOrSchedule, eps: float = 1e-5
) -> base.GradientTransformation:
    """The online newton optimizer.

    It extends optax optimizers.

    References
    ----------
    [^1] [Hazan, E., Agarwal, A. and Kale, S., 2007. Logarithmic regret algorithms for
    online convex optimization. Machine Learning, 69(2-3), pp.169-192]
    (https://link.springer.com/content/pdf/10.1007/s10994-007-5016-8.pdf)

    Args:
      learning_rate: this is a fixed global scaling factor.
      initial_accumulator_value: initialisation for the accumulator.
      eps: a small constant applied to denominator inside of the square root
        (as in RMSProp) to avoid dividing by zero when rescaling.

    Returns:
      the corresponding `GradientTransformation`.
    """
    return combine.chain(
        scale_by_newton(eps=eps),
        _scale_by_learning_rate(learning_rate),
    )


class ScaleByNewtonState(base.OptState):
    """State holding the sum of gradient squares to date."""

    hessian_inv: base.Updates


def sherman_morrison(A_inv, u, v):
    den = 1.0 + (u @ A_inv) @ v
    A_inv -= A_inv @ jnp.outer(u, v) @ A_inv / den
    return A_inv


def scale_by_newton(eps: float = 1e-7) -> base.GradientTransformation:
    """Rescale updates by multiplying by the inverse of an approximation of the hessian.

    References
    ----------
    [^1]: [Hazan, E., Agarwal, A. and Kale, S., 2007. Logarithmic regret algorithms for online convex optimization. Machine Learning, 69(2-3), pp.169-192](https://www.cs.princeton.edu/~ehazan/papers/log-journal.pdf)

    Args:
      eps: A small floating point value to avoid zero denominator.

    Returns:
      An (init_fn, update_fn) tuple.
    """

    def init_fn(params):
        hessian_inv = jax.tree_map(
            lambda t: jnp.eye(len(t.flatten()), dtype=t.dtype) / eps, params
        )
        return ScaleByNewtonState(hessian_inv=hessian_inv)

    def update_fn(updates, state, params=None):
        del params

        class Tuple(tuple):
            """Class to avoid pytree conversion and allow for the use
            of shapes in final reshape."""

            ...

        shapes = jax.tree_map(lambda x: Tuple(x.shape), updates)
        updates = jax.tree_map(lambda x: x.flatten(), updates)
        hessian_inv = jax.tree_multimap(
            lambda u, hinv: sherman_morrison(hinv, u, u), updates, state.hessian_inv
        )
        updates = jax.tree_multimap(lambda hinv, g: hinv @ g, hessian_inv, updates)
        updates = jax.tree_multimap(lambda u, shape: u.reshape(shape), updates, shapes)

        return updates, ScaleByNewtonState(hessian_inv=hessian_inv)

    return base.GradientTransformation(init_fn, update_fn)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from wax.gym.callbacks.callbacks import Callback, normalize_callback, unpack_callbacks


def test_callback():
    cb = Callback()
    assert isinstance(cb, Callback)

    cb = Callback(on_train_start=True)
    assert isinstance(cb, Callback)

    cb = Callback(on_act=True)
    assert isinstance(cb, Callback)

    cb = Callback(on_step=True)
    assert isinstance(cb, Callback)

    cb = Callback(on_train_end=True)
    assert isinstance(cb, Callback)

    cb.register()
    assert isinstance(cb, Callback)

    cb.unregister()
    assert isinstance(cb, Callback)
    _, _, _, _ = unpack_callbacks(None)
    ts, a, s, te = unpack_callbacks([[None, None, None, cb]])
    assert te[0] is cb
    with pytest.raises(TypeError):
        normalize_callback(None)
    ts, a, s, te = normalize_callback(cb)
    assert te is True
    cbn = normalize_callback((cb,))
    assert cbn == (cb,)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
from jax import numpy as jnp

from wax.modules.arma import ARMA
from wax.unroll import unroll_transform_with_state


def test_arma():
    alpha = jnp.array([0.6, -0.5, 0.4, -0.4, 0.3])
    beta = jnp.array([0.3, -0.2])

    rng = jax.random.PRNGKey(42)
    eps = jax.random.normal(rng, (100,))
    sim = unroll_transform_with_state(lambda eps: ARMA(alpha, beta)(eps))
    params, state = sim.init(rng, eps)
    y, state = sim.apply(params, state, rng, eps)
    assert y.shape == (100,)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from importlib import import_module
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
from typing_extensions import Literal

from .._index import index
from ..types import Axes, AxisAxes, Shape, ShapeOrScalar
from .base import BaseTensor, unwrap1, unwrap_
from .tensor import Tensor, TensorOrScalar, TensorType

if TYPE_CHECKING:
    import tensorflow as tf  # for static analyzers

    from .extensions import NormsMethods  # noqa: F401
else:
    # lazy import in TensorFlowTensor
    tf = None

FuncType = Callable[..., Any]
F = TypeVar("F", bound=FuncType)


def samedevice(f: F) -> F:
    @functools.wraps(f)
    def wrapper(self: "TensorFlowTensor", *args: Any, **kwargs: Any) -> Any:
        with tf.device(self.raw.device):
            return f(self, *args, **kwargs)

    return cast(F, wrapper)


def common_dtype(f: F) -> F:
    @functools.wraps(f)
    def wrapper(self: "TensorFlowTensor", *args: Any, **kwargs: Any) -> Any:
        dtypes = {self.dtype} | {arg.dtype for arg in args if isinstance(arg, Tensor)}
        if len(dtypes) == 1:
            # all dtypes are the same, nothing more to do
            return f(self, *args, **kwargs)
        numpy_dtypes = [np.dtype(dtype.name) for dtype in dtypes]
        common = np.find_common_type(numpy_dtypes, [])
        common = getattr(tf, common.name)
        if self.dtype != common:
            self = self.astype(common)
        args = tuple(
            arg.astype(common)
            if isinstance(arg, Tensor) and arg.dtype != common
            else arg
            for arg in args
        )
        return f(self, *args, **kwargs)

    return cast(F, wrapper)


def assert_bool(x: Any) -> None:
    if not isinstance(x, Tensor):
        return
    if x.dtype != tf.bool:
        raise ValueError(f"requires dtype bool, got {x.dtype}, consider t.bool().all()")


class TensorFlowTensor(BaseTensor):
    __slots__ = ()

    # more specific types for the extensions
    norms: "NormsMethods[TensorFlowTensor]"

    def __init__(self, raw: "tf.Tensor"):  # type: ignore
        global tf
        if tf is None:
            tf = import_module("tensorflow")
        super().__init__(raw)

    @property
    def raw(self) -> "tf.Tensor":  # type: ignore
        return super().raw

    def numpy(self: TensorType) -> Any:
        a = self.raw.numpy()
        if a.flags.writeable:
            # without the check, we would attempt to set it on array
            # scalars, and that would fail
            a.flags.writeable = False
        return a

    def item(self: TensorType) -> Union[int, float, bool]:
        return self.numpy().item()  # type: ignore

    @property
    def shape(self: TensorType) -> Shape:
        return tuple(self.raw.shape.as_list())

    def reshape(self: TensorType, shape: Union[Shape, int]) -> TensorType:
        if isinstance(shape, int):
            shape = (shape,)
        return type(self)(tf.reshape(self.raw, shape))

    def astype(self: TensorType, dtype: Any) -> TensorType:
        return type(self)(tf.cast(self.raw, dtype))

    def clip(self: TensorType, min_: float, max_: float) -> TensorType:
        return type(self)(tf.clip_by_value(self.raw, min_, max_))

    def square(self: TensorType) -> TensorType:
        return type(self)(tf.square(self.raw))

    def arctanh(self: TensorType) -> TensorType:
        return type(self)(tf.atanh(self.raw))

    def sum(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        if self.raw.dtype == tf.bool:
            return self.astype(tf.int64).sum(axis=axis, keepdims=keepdims)
        return type(self)(tf.reduce_sum(self.raw, axis=axis, keepdims=keepdims))

    def prod(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        if self.raw.dtype == tf.bool:
            return self.astype(tf.int64).prod(axis=axis, keepdims=keepdims)
        return type(self)(tf.reduce_prod(self.raw, axis=axis, keepdims=keepdims))

    def mean(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        if self.raw.dtype not in [tf.float16, tf.float32, tf.float64]:
            raise ValueError(
                f"Can only calculate the mean of floating types. Got {self.raw.dtype} instead."
            )
        return type(self)(tf.reduce_mean(self.raw, axis=axis, keepdims=keepdims))

    def min(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        return type(self)(tf.reduce_min(self.raw, axis=axis, keepdims=keepdims))

    def max(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        return type(self)(tf.reduce_max(self.raw, axis=axis, keepdims=keepdims))

    def minimum(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(tf.minimum(self.raw, unwrap1(other)))

    def maximum(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(tf.maximum(self.raw, unwrap1(other)))

    def argmin(self: TensorType, axis: Optional[int] = None) -> TensorType:
        return type(self)(tf.argmin(self.raw, axis=axis))

    def argmax(self: TensorType, axis: Optional[int] = None) -> TensorType:
        return type(self)(tf.argmax(self.raw, axis=axis))

    def argsort(self: TensorType, axis: Optional[int] = -1) -> TensorType:
        return type(self)(tf.argsort(self.raw, axis=axis))

    def sort(self: TensorType, axis: Optional[int] = -1) -> TensorType:
        return type(self)(tf.sort(self.raw, axis=axis))

    def topk(
        self: TensorType, k: int, sorted: bool = True
    ) -> Tuple[TensorType, TensorType]:
        values, indices = tf.math.top_k(self.raw, k, sorted=sorted)
        return type(self)(values), type(self)(indices)

    @samedevice
    def uniform(
        self: TensorType, shape: ShapeOrScalar, low: float = 0.0, high: float = 1.0
    ) -> TensorType:
        if not isinstance(shape, Iterable):
            shape = (shape,)
        return type(self)(
            tf.random.uniform(shape, minval=low, maxval=high, dtype=self.raw.dtype)
        )

    @samedevice
    def normal(
        self: TensorType, shape: ShapeOrScalar, mean: float = 0.0, stddev: float = 1.0
    ) -> TensorType:
        if not isinstance(shape, Iterable):
            shape = (shape,)
        return type(self)(
            tf.random.normal(shape, mean=mean, stddev=stddev, dtype=self.raw.dtype)
        )

    @samedevice
    def ones(self: TensorType, shape: ShapeOrScalar) -> TensorType:
        return type(self)(tf.ones(shape, dtype=self.raw.dtype))

    @samedevice
    def zeros(self: TensorType, shape: ShapeOrScalar) -> TensorType:
        return type(self)(tf.zeros(shape, dtype=self.raw.dtype))

    def ones_like(self: TensorType) -> TensorType:
        return type(self)(tf.ones_like(self.raw))

    def zeros_like(self: TensorType) -> TensorType:
        return type(self)(tf.zeros_like(self.raw))

    def full_like(self: TensorType, fill_value: float) -> TensorType:
        fill_value = tf.cast(fill_value, self.raw.dtype)
        return type(self)(tf.fill(self.raw.shape, fill_value))

    @samedevice
    def onehot_like(
        self: TensorType, indices: TensorType, *, value: float = 1
    ) -> TensorType:
        if self.ndim != 2:
            raise ValueError("onehot_like only supported for 2D tensors")
        if indices.ndim != 1:
            raise ValueError("onehot_like requires 1D indices")
        if len(indices) != len(self):
            raise ValueError("length of indices must match length of tensor")
        value = tf.cast(value, self.raw.dtype)
        return type(self)(
            tf.one_hot(
                indices.raw,
                depth=self.raw.shape[-1],
                on_value=value,
                dtype=self.raw.dtype,
            )
        )

    @samedevice
    def from_numpy(self: TensorType, a: Any) -> TensorType:
        return type(self)(tf.convert_to_tensor(a))

    def _concatenate(
        self: TensorType, tensors: Iterable[TensorType], axis: int = 0
    ) -> TensorType:
        # concatenates only "tensors", but not "self"
        tensors_ = unwrap_(*tensors)
        return type(self)(tf.concat(tensors_, axis=axis))

    def _stack(
        self: TensorType, tensors: Iterable[TensorType], axis: int = 0
    ) -> TensorType:
        # stacks only "tensors", but not "self"
        tensors_ = unwrap_(*tensors)
        return type(self)(tf.stack(tensors_, axis=axis))

    def transpose(self: TensorType, axes: Optional[Axes] = None) -> TensorType:
        if axes is None:
            axes = tuple(range(self.ndim - 1, -1, -1))
        return type(self)(tf.transpose(self.raw, perm=axes))

    def all(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        assert_bool(self)
        return type(self)(tf.reduce_all(self.raw, axis=axis, keepdims=keepdims))

    def any(
        self: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
    ) -> TensorType:
        assert_bool(self)
        return type(self)(tf.reduce_any(self.raw, axis=axis, keepdims=keepdims))

    def logical_and(self: TensorType, other: TensorOrScalar) -> TensorType:
        assert_bool(self)
        assert_bool(other)
        return type(self)(tf.logical_and(self.raw, unwrap1(other)))

    def logical_or(self: TensorType, other: TensorOrScalar) -> TensorType:
        assert_bool(self)
        assert_bool(other)
        return type(self)(tf.logical_or(self.raw, unwrap1(other)))

    def logical_not(self: TensorType) -> TensorType:
        assert_bool(self)
        return type(self)(tf.logical_not(self.raw))

    def exp(self: TensorType) -> TensorType:
        return type(self)(tf.exp(self.raw))

    def log(self: TensorType) -> TensorType:
        return type(self)(tf.math.log(self.raw))

    def log2(self: TensorType) -> TensorType:
        return type(self)(tf.math.log(self.raw) / tf.math.log(2.0))

    def log10(self: TensorType) -> TensorType:
        return type(self)(tf.math.log(self.raw) / tf.math.log(10.0))

    def log1p(self: TensorType) -> TensorType:
        return type(self)(tf.math.log1p(self.raw))

    def tile(self: TensorType, multiples: Axes) -> TensorType:
        multiples = unwrap1(multiples)
        if len(multiples) != self.ndim:
            raise ValueError("multiples requires one entry for each dimension")
        return type(self)(tf.tile(self.raw, multiples))

    def softmax(self: TensorType, axis: int = -1) -> TensorType:
        return type(self)(tf.nn.softmax(self.raw, axis=axis))

    def log_softmax(self: TensorType, axis: int = -1) -> TensorType:
        return type(self)(tf.nn.log_softmax(self.raw, axis=axis))

    def squeeze(self: TensorType, axis: Optional[AxisAxes] = None) -> TensorType:
        return type(self)(tf.squeeze(self.raw, axis=axis))

    def expand_dims(self: TensorType, axis: int) -> TensorType:
        return type(self)(tf.expand_dims(self.raw, axis=axis))

    @samedevice
    def full(self: TensorType, shape: ShapeOrScalar, value: float) -> TensorType:
        if not isinstance(shape, Iterable):
            shape = (shape,)
        return type(self)(tf.fill(shape, value))

    def index_update(
        self: TensorType, indices: Any, values: TensorOrScalar
    ) -> TensorType:
        indices, values_ = unwrap_(indices, values)
        del values
        if isinstance(indices, tuple):
            indices = unwrap_(*indices)

        x = self.raw
        if isinstance(indices, int):
            indices = tf.convert_to_tensor(indices)

        if isinstance(indices, tf.Tensor):
            if isinstance(values_, int) or isinstance(values_, float):
                values_ = tf.fill(x.shape[-1:], values_)
            return type(self)(
                tf.tensor_scatter_nd_update(x, [[indices]], values_[None])
            )
        elif isinstance(indices, tuple) and any(
            isinstance(idx, slice) for idx in indices
        ):
            if (
                len(indices) == x.ndim == 2
                and indices[0] == index[:]
                and not isinstance(indices[1], slice)
            ):
                x = tf.transpose(x)
                if isinstance(values_, int) or isinstance(values_, float):
                    values_ = tf.fill(x.shape[-1:], values_)
                result = tf.tensor_scatter_nd_update(x, [[indices[-1]]], values_[None])
                return type(self)(tf.transpose(result))
            else:
                raise NotImplementedError  # pragma: no cover
        elif isinstance(indices, tuple):
            if all(idx.dtype in [tf.int32, tf.int64] for idx in indices):
                indices = [
                    tf.cast(idx, tf.int64) if idx.dtype == tf.int32 else idx
                    for idx in indices
                ]
            indices = tf.stack(indices, axis=-1)
            if isinstance(values_, int) or isinstance(values_, float):
                values_ = tf.fill((indices.shape[0],), values_)
            return type(self)(tf.tensor_scatter_nd_update(x, indices, values_))
        else:
            raise ValueError  # pragma: no cover

    @samedevice
    def arange(
        self: TensorType,
        start: int,
        stop: Optional[int] = None,
        step: Optional[int] = None,
    ) -> TensorType:
        if step is None:
            step = 1
        if stop is None:
            stop = start
            start = 0
        return type(self)(tf.range(start, stop, step))

    def cumsum(self: TensorType, axis: Optional[int] = None) -> TensorType:
        if axis is None:
            x = tf.reshape(self.raw, (-1,))
            return type(self)(tf.cumsum(x, axis=0))
        return type(self)(tf.cumsum(self.raw, axis=axis))

    def flip(self: TensorType, axis: Optional[AxisAxes] = None) -> TensorType:
        if axis is None:
            axis = tuple(range(self.ndim))
        if not isinstance(axis, Iterable):
            axis = (axis,)
        return type(self)(tf.reverse(self.raw, axis=axis))

    def meshgrid(
        self: TensorType, *tensors: TensorType, indexing: str = "xy"
    ) -> Tuple[TensorType, ...]:
        tensors = unwrap_(*tensors)
        outputs = tf.meshgrid(self.raw, *tensors, indexing=indexing)
        return tuple(type(self)(out) for out in outputs)

    def pad(
        self: TensorType,
        paddings: Tuple[Tuple[int, int], ...],
        mode: str = "constant",
        value: float = 0,
    ) -> TensorType:
        if len(paddings) != self.ndim:
            raise ValueError("pad requires a tuple for each dimension")
        for p in paddings:
            if len(p) != 2:
                raise ValueError("pad requires a tuple for each dimension")
        if not (mode == "constant" or mode == "reflect"):
            raise ValueError("pad requires mode 'constant' or 'reflect'")
        if mode == "reflect":
            # PyTorch's pad has limited support for 'reflect' padding
            if self.ndim != 3 and self.ndim != 4:
                raise NotImplementedError  # pragma: no cover
            k = self.ndim - 2
            if paddings[:k] != ((0, 0),) * k:
                raise NotImplementedError  # pragma: no cover
        return type(self)(tf.pad(self.raw, paddings, mode=mode, constant_values=value))

    def isnan(self: TensorType) -> TensorType:
        return type(self)(tf.math.is_nan(self.raw))

    def isinf(self: TensorType) -> TensorType:
        return type(self)(tf.math.is_inf(self.raw))

    def crossentropy(self: TensorType, labels: TensorType) -> TensorType:
        if self.ndim != 2:
            raise ValueError("crossentropy only supported for 2D logits tensors")
        if self.shape[:1] != labels.shape:
            raise ValueError("labels must be 1D and must match the length of logits")
        return type(self)(
            tf.nn.sparse_softmax_cross_entropy_with_logits(labels.raw, self.raw)
        )

    def slogdet(self: TensorType) -> Tuple[TensorType, TensorType]:
        sign, logabsdet = tf.linalg.slogdet(self.raw)
        return type(self)(sign), type(self)(logabsdet)

    @overload
    def _value_and_grad_fn(
        self: TensorType, f: Callable[..., TensorType]
    ) -> Callable[..., Tuple[TensorType, TensorType]]:
        ...

    @overload  # noqa: F811 (waiting for pyflakes > 2.1.1)
    def _value_and_grad_fn(
        self: TensorType, f: Callable[..., TensorType], has_aux: Literal[False]
    ) -> Callable[..., Tuple[TensorType, TensorType]]:
        ...

    @overload  # noqa: F811 (waiting for pyflakes > 2.1.1)
    def _value_and_grad_fn(
        self: TensorType,
        f: Callable[..., Tuple[TensorType, Any]],
        has_aux: Literal[True],
    ) -> Callable[..., Tuple[TensorType, Any, TensorType]]:
        ...

    def _value_and_grad_fn(  # noqa: F811 (waiting for pyflakes > 2.1.1)
        self: TensorType, f: Callable, has_aux: bool = False
    ) -> Callable[..., Tuple]:
        def value_and_grad(x: TensorType, *args: Any, **kwargs: Any) -> Tuple:
            # using tf.identity to make x independent from possible other instances of x in args
            x_ = TensorFlowTensor(tf.identity(x.raw))
            del x
            with tf.GradientTape() as tape:
                tape.watch(x_.raw)
                if has_aux:
                    loss, aux = f(x_, *args, **kwargs)
                else:
                    loss = f(x_, *args, **kwargs)
            grad = tape.gradient(loss.raw, x_.raw)
            grad = TensorFlowTensor(grad)
            assert grad.shape == x_.shape
            if has_aux:
                return loss, aux, grad
            else:
                return loss, grad

        return value_and_grad

    def sign(self: TensorType) -> TensorType:
        return type(self)(tf.sign(self.raw))

    def sqrt(self: TensorType) -> TensorType:
        return type(self)(tf.sqrt(self.raw))

    def tanh(self: TensorType) -> TensorType:
        return type(self)(tf.tanh(self.raw))

    def float32(self: TensorType) -> TensorType:
        return self.astype(tf.float32)

    def float64(self: TensorType) -> TensorType:
        return self.astype(tf.float64)

    def where(self: TensorType, x: TensorOrScalar, y: TensorOrScalar) -> TensorType:
        x, y = unwrap_(x, y)
        return type(self)(tf.where(self.raw, x, y))

    @common_dtype
    def __lt__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__lt__(unwrap1(other)))

    @common_dtype
    def __le__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__le__(unwrap1(other)))

    @common_dtype
    def __eq__(self: TensorType, other: TensorOrScalar) -> TensorType:  # type: ignore
        return type(self)(self.raw.__eq__(unwrap1(other)))

    @common_dtype
    def __ne__(self: TensorType, other: TensorOrScalar) -> TensorType:  # type: ignore
        return type(self)(self.raw.__ne__(unwrap1(other)))

    @common_dtype
    def __gt__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__gt__(unwrap1(other)))

    @common_dtype
    def __ge__(self: TensorType, other: TensorOrScalar) -> TensorType:
        return type(self)(self.raw.__ge__(unwrap1(other)))

    def __getitem__(self: TensorType, index: Any) -> TensorType:
        if isinstance(index, tuple):
            index = tuple(x.raw if isinstance(x, Tensor) else x for x in index)
            basic = all(
                x is None or x is Ellipsis or isinstance(x, int) or isinstance(x, slice)
                for x in index
            )
            if not basic:
                # workaround for missing support for this in TensorFlow
                index = [tf.convert_to_tensor(x) for x in index]
                shapes = [tuple(x.shape) for x in index]
                shape = tuple(max(x) for x in zip(*shapes))
                int64 = any(x.dtype == tf.int64 for x in index)
                for i in range(len(index)):
                    t = index[i]
                    if int64:
                        t = tf.cast(t, tf.int64)  # pragma: no cover
                    assert t.ndim == len(shape)
                    tiling = []
                    for b, k in zip(shape, t.shape):
                        if k == 1:
                            tiling.append(b)
                        elif k == b:
                            tiling.append(1)
                        else:
                            raise ValueError(  # pragma: no cover
                                f"{tuple(t.shape)} cannot be broadcasted to {shape}"
                            )
                    index[i] = tf.tile(t, tiling)
                index = tf.stack(index, axis=-1)
                return type(self)(tf.gather_nd(self.raw, index))
        elif (
            isinstance(index, range)
            or isinstance(index, list)
            or isinstance(index, np.ndarray)
        ):
            return type(self)(tf.gather(self.raw, index))
        elif isinstance(index, Tensor):
            if index.raw.dtype == tf.bool:
                return type(self)(self.raw.__getitem__(index.raw))
            else:
                return type(self)(tf.gather(self.raw, index.raw))
        return type(self)(self.raw.__getitem__(index))

    def take_along_axis(self: TensorType, index: TensorType, axis: int) -> TensorType:
        axis = batch_dims = axis % self.ndim
        if axis != self.ndim - 1:
            raise NotImplementedError(
                "take_along_axis is currently only supported for the last axis"
            )
        return type(self)(
            tf.gather(self.raw, index.raw, axis=axis, batch_dims=batch_dims)
        )

    def bool(self: TensorType) -> TensorType:
        return self.astype(tf.bool)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax.numpy as jnp
import numpy as onp
import pandas as pd
import pytest
import xarray as xr
from jax.config import config

from wax.accessors import register_wax_accessors
from wax.datasets.generate_temperature_data import generate_temperature_data
from wax.format import format_dataframe, format_series
from wax.modules import EWMA
from wax.stream_test import prepare_test_data
from wax.testing import assert_tree_all_close
from wax.unroll import unroll


def module_same_shape(x):
    return EWMA(1.0 / 10.0)(x)


def module_reduce_shape(x):
    return EWMA(1.0 / 10.0)(x).sum()


def module_map(x):
    return {"ts_10": EWMA(1.0 / 10.0)(x).sum(), "ts_5": EWMA(1.0 / 5.0)(x).sum()}


def check_ema_state(state, ref_count=124):
    assert (state["ewma"]["count"] == ref_count).all()


def prepare_format_data(format):
    dataset = prepare_test_data()

    if format == "dataset":
        return dataset
    data = dataset["ground"]

    if format == "dataframe":
        # simple index dataframe
        # data = data.isel(lon=0).to_pandas()
        data = data.to_series().unstack()
    elif format == "series":
        data = data.to_series()
    else:
        assert format == "dataarray"
    return data


def test_unroll_dataset_accessor():
    config.update("jax_enable_x64", False)
    register_wax_accessors()

    dataset = prepare_format_data("dataset")

    def module(dataset):
        return EWMA(1.0 / 10.0)(dataset["ground"])

    with pytest.raises(ValueError):
        dataset.wax.stream().apply(
            module, format_dims=onp.array(["time", "lon", "lat"])
        )

    xoutput, state = dataset.wax.stream(
        local_time="time",
        freqs={"day": "d"},
        ffills={"day": True},
        buffer_maxlen={"news_time": 3},
        verbose=False,
        pbar=False,
    ).apply(module, format_dims=onp.array(["time", "lon", "lat"]))
    assert xoutput.shape == (124, 53, 25)
    assert isinstance(xoutput, xr.DataArray)
    check_ema_state(state)


@pytest.mark.parametrize("format", ["dataarray", "dataframe", "series"])
def test_unroll_dataarray_accessor_module_same_shape(format):
    config.update("jax_enable_x64", False)
    module = module_same_shape
    register_wax_accessors()
    data = prepare_format_data(format)
    if format == "dataframe":
        assert data.index.nlevels == 2
        assert data.columns.nlevels == 1
    xoutput, state = data.wax.stream().apply(module)
    assert type(xoutput) == type(data)
    assert xoutput.shape == data.shape
    check_ema_state(state)


@pytest.mark.parametrize("format", ["dataframe"])
def test_unroll_dataframe_transposed_accessor_module(format):
    config.update("jax_enable_x64", False)
    module = module_same_shape
    register_wax_accessors()
    data = prepare_format_data(format).unstack()
    if format == "dataframe":
        assert data.index.nlevels == 1
        assert data.columns.nlevels == 2
    xoutput, state = data.wax.stream().apply(module)
    assert type(xoutput) == type(data)
    assert xoutput.shape == data.shape  # (124, 53, 25)
    check_ema_state(state)


def test_unroll_dataframe_accessor_simple_index():
    config.update("jax_enable_x64", False)
    module = module_same_shape
    register_wax_accessors()
    data = prepare_format_data("dataarray")
    data = data.isel(lon=0).to_pandas()
    assert data.index.nlevels == 1
    assert data.columns.nlevels == 1

    xoutput, state = data.wax.stream().apply(module)

    assert type(xoutput) == type(data)

    assert xoutput.shape == data.shape
    check_ema_state(state)


@pytest.mark.parametrize("format", ["dataarray", "dataframe", "series"])
def test_unroll_dataarray_accessor_module_reduce_shape(format):
    config.update("jax_enable_x64", False)
    module = module_reduce_shape
    register_wax_accessors()
    data = prepare_format_data(format)
    xoutput, state = data.wax.stream().apply(module)

    if format in ["series", "dataarray"]:
        assert type(xoutput) == type(data)
    else:
        assert isinstance(xoutput, pd.Series)

    assert xoutput.shape == (124,)
    check_ema_state(state)


@pytest.mark.parametrize("format", ["dataarray", "dataframe", "series"])
def test_unroll_dataarray_accessor_module_map(format):
    config.update("jax_enable_x64", False)
    module = module_map
    register_wax_accessors()
    data = prepare_format_data(format)
    xoutput, state = data.wax.stream().apply(module)

    assert isinstance(xoutput, dict)
    assert xoutput.keys() == {"ts_10", "ts_5"}

    check_ema_state(state, ref_count=124)


def test_unroll_dataset_accessor_module_map():
    config.update("jax_enable_x64", False)

    def module_map(dataset):
        return {
            name: {
                "ts_10": EWMA(1.0 / 10.0, name=name)(val).sum(),
                "ts_5": EWMA(1.0 / 5.0, name=name)(val).sum(),
            }
            for name, val in dataset.items()
            if name in ["ground", "air"]
        }

    module = module_map
    register_wax_accessors()
    dataset = prepare_test_data()
    xoutput, state = dataset.wax.stream(local_time="time").apply(
        module, format_dims=["time"]
    )
    assert isinstance(xoutput, dict)
    assert xoutput.keys() == {"air", "ground"}
    assert state.keys() == {"air", "air_1", "ground", "ground_1"}
    check_ema_state({"ewma": state["ground"]}, ref_count=124)
    check_ema_state({"ewma": state["ground_1"]}, ref_count=124)
    check_ema_state({"ewma": state["air"]}, ref_count=6)
    check_ema_state({"ewma": state["air_1"]}, ref_count=6)


@pytest.mark.parametrize("format", ["dataarray", "dataframe", "series"])
def test_wax_ewma(format):
    config.update("jax_enable_x64", False)
    register_wax_accessors()
    data = prepare_format_data(format)
    ema, state = data.wax.ewm(alpha=0.1, adjust=True, return_state=True).mean()
    check_ema_state(state, ref_count=124)


def _compute_ewma_direct(dataarray):

    x = jnp.array(dataarray.values, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EWMA(0.1, adjust=True)(x)

    ema3, state3 = unroll(model, return_final_state=True)(x)
    return ema3, state3


@pytest.mark.parametrize("format", ["dataframe"])
def test_wax_ewma_vs_pandas(format):

    config.update("jax_enable_x64", True)
    register_wax_accessors()
    data = prepare_format_data(format)

    pandas_ema = data.unstack().ewm(alpha=0.1, adjust=True).mean().stack()
    ema = data.wax.ewm(alpha=0.1, adjust=True).mean()
    ema2, state = data.wax.ewm(alpha=0.1, adjust=True, return_state=True).mean()

    du = data.unstack()
    ema3, state3 = _compute_ewma_direct(du)
    ema3 = pd.DataFrame(ema3, index=du.index, columns=du.columns).stack().values

    assert onp.allclose(ema2, pandas_ema.values)
    assert onp.allclose(ema3, pandas_ema.values)
    assert onp.allclose(ema3, ema.values)
    check_ema_state(state, ref_count=124)


def test_air_temperature_dataset():
    config.update("jax_enable_x64", True)
    register_wax_accessors()
    # da = xr.tutorial.open_dataset("air_temperature")
    da = generate_temperature_data()

    def diff(y):
        return y["air"].flatten().sum()

    output = da.wax.stream(return_state=False).apply(diff, format_dims=da.air.dims)
    check = da.air.to_series().sum(level=0)
    assert_tree_all_close(output.to_series(), check)


def test_air_temperature_dataarray():
    config.update("jax_enable_x64", True)
    register_wax_accessors()
    # da = xr.tutorial.open_dataset("air_temperature")
    da = generate_temperature_data()
    da = da.air

    def diff(y):
        return y.flatten().sum()

    output = da.wax.stream(return_state=False).apply(diff, format_dims=da.dims)
    check = da.to_series().sum(level=0)
    assert_tree_all_close(output.to_series(), check)


def test_air_temperature_dataframe():
    config.update("jax_enable_x64", True)
    register_wax_accessors()
    # da = xr.tutorial.open_dataset("air_temperature")
    da = generate_temperature_data()
    da = da.air.isel(lat=0)

    def diff(y):
        return y.flatten().sum()

    output = (
        da.to_pandas().wax.stream(return_state=False).apply(diff, format_dims=da.dims)
    )
    check = da.to_series().sum(level=0)
    assert_tree_all_close(output, check)


def test_air_temperature_series():
    config.update("jax_enable_x64", True)
    register_wax_accessors()
    # da = xr.tutorial.open_dataset("air_temperature")
    dataset = generate_temperature_data()
    dataset = dataset.air

    def diff(y):
        return y.flatten().sum()

    output = (
        dataset.to_series()
        .wax.stream(return_state=False)
        .apply(diff, format_dims=dataset.dims)
    )
    check = dataset.to_series().sum(level=0)
    assert_tree_all_close(output, check)


def test_ewm_dataframe():
    register_wax_accessors()
    T = int(1.0e2)
    N = 10
    dataframe = pd.DataFrame(
        onp.random.normal(size=(T, N)), index=pd.date_range("1970", periods=T)
    )
    y = dataframe.ewm(alpha=1.0 / 10.0).mean()
    y2 = dataframe.wax.ewm(alpha=1.0 / 10.0).mean()
    assert (y - y2).abs().stack().max() < 1.0e-5


def test_ewm_dataframe_no_format_outputs():
    register_wax_accessors()

    T = int(1.0e3)
    N = 10
    dataframe = pd.DataFrame(
        onp.random.normal(size=(T, N)), index=pd.date_range("1970", periods=T)
    )
    y = dataframe.ewm(alpha=1.0 / 10.0).mean()
    y2 = dataframe.wax.ewm(alpha=1.0 / 10.0, format_outputs=False).mean()
    y2 = pd.DataFrame(onp.array(y2), index=y.index)

    assert (y - y2).abs().stack().max() < 1.0e-5

    y2 = dataframe.wax.ewm(alpha=1.0 / 10.0).mean()

    dataset = xr.DataArray(y2)
    output = y2.values

    # format dataframe
    # with no format_dims
    output_dataframe = format_dataframe(dataset.coords, output)
    assert isinstance(output_dataframe, pd.DataFrame)
    assert (output_dataframe.values == output).flatten().all()

    # with empty coords
    output_dataframe = format_dataframe(None, output)
    assert isinstance(output_dataframe, pd.DataFrame)
    assert (output_dataframe.values == output).flatten().all()

    # with explicit format_dims, but with nested data structure.
    # test when a nested data structure is passed to format_dataframe.
    outputs = {"output": output}
    format_dims = {"output": onp.array(dataset.dims)}
    outputs_dataframe = format_dataframe(dataset.coords, outputs, format_dims)
    assert isinstance(outputs_dataframe["output"], pd.DataFrame)
    assert (outputs_dataframe["output"].values == outputs["output"]).flatten().all()

    # format series
    # with explicit format_dims, but with nested data structure.
    # test when a nested data structure is passed to format_dataframe.
    outputs = {"output": output.sum(1)}
    format_dims = {"output": onp.array(dataset.dims)[0]}
    outputs_dataframe = format_series(dataset.coords, outputs, format_dims)
    assert isinstance(outputs_dataframe["output"], pd.Series)
    assert (outputs_dataframe["output"].values == outputs["output"]).flatten().all()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compute exponentioal moving average."""
import haiku as hk
import jax.numpy as jnp


class EWMA(hk.Module):
    """Compute exponentioal moving average."""

    def __init__(
        self, alpha: float, adjust: bool = True, initial_value=jnp.nan, name: str = None
    ):
        """Initialize module.

        Args:
            alpha: alpha parameter of the exponential moving average.
            adjust: if true, implement a non-stationary filter with exponential initialization
                scheme. If "linear", implement a non-stationary filter with linear initialization.
            initial_value: initial value for the state.
            name : name of the module instance.
        """
        super().__init__(name=name)
        self.alpha = alpha
        self.adjust = adjust
        self.initial_value = initial_value

    def __call__(self, x):
        """Compute EWMA.

        Args:
            x: input data.
        """
        mean = hk.get_state(
            "mean",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, self.initial_value, dtype),
        )
        is_initialized = hk.get_state(
            "is_initialized",
            shape=x.shape,
            dtype=bool,
            init=lambda shape, dtype: jnp.full(shape, False, dtype),
        )
        count = hk.get_state(
            "count",
            shape=x.shape,
            dtype=x.dtype,
            init=lambda shape, dtype: jnp.full(shape, 0.0, dtype),
        )

        alpha = hk.get_parameter(
            "alpha",
            shape=[],
            dtype=x.dtype,
            init=lambda *_: jnp.array(self.alpha),
        )

        # initialization on first non-nan value
        mean = jnp.where(is_initialized, mean, x)
        is_initialized = jnp.where(jnp.isnan(x), is_initialized, True)
        hk.set_state("is_initialized", is_initialized)

        isnan_x = jnp.isnan(x)
        x = jnp.nan_to_num(x)
        mean = jnp.nan_to_num(mean)

        count = jnp.where(isnan_x, count, count + 1)

        # alpha adjustement scheme
        if self.adjust == "linear":
            tscale = 1.0 / alpha
            tscale = jnp.where(count < tscale, count, tscale)
            alpha = jnp.where(tscale > 0, 1.0 / tscale, jnp.nan)
        elif self.adjust:
            # exponential scheme (as in pandas)
            eps = jnp.finfo(x.dtype).resolution
            alpha = jnp.where(
                count > 0, alpha / (eps + 1.0 - (1.0 - alpha) ** (count)), 1.0
            )
            # alpha = alpha / (1.0 - (1.0 - alpha) ** (count))

        # update mean  if x is not nan
        mean = jnp.where(isnan_x, mean, (1.0 - alpha) * mean + alpha * x)

        mean = jnp.where(is_initialized, mean, self.initial_value)
        hk.set_state("mean", mean)
        hk.set_state("count", count)
        return mean



# Copyright 2014-2019, xarray Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate fake temperature data for tests purposes."""
import numpy as onp
import pandas as pd
import xarray as xr


def get_lats_lons_times(n_lat=25, n_lon=53, n_time=124, freq="H"):
    lats = onp.linspace(0, 20, n_lat)
    lons = onp.linspace(0, 11, n_lon)
    times = pd.date_range(start="2000/01/01", freq=freq, periods=n_time)
    return lats, lons, times


def generate_data(n_lat=25, n_lon=53, n_time=124, freq="H"):
    onp.random.seed(3)
    lats, lons, times = get_lats_lons_times(n_lat, n_lon, n_time, freq=freq)

    data = onp.random.rand(n_lon, n_lat, n_time)

    data = xr.DataArray(
        data,
        coords=[lons, lats, times],
        dims=["lon", "lat", "time"],
        name="data",
    ).transpose("time", "lon", "lat")
    return data


def generate_times_data(n_lat=25, n_lon=53, n_time=124, freq="H"):
    onp.random.seed(3)

    lats, lons, times = get_lats_lons_times(n_lat, n_lon, n_time, freq=freq)

    times_arr = onp.random.choice(times, size=(n_lon, n_lat, n_time))
    times_data = xr.DataArray(
        times_arr,
        coords=[lons, lats, times],
        dims=["lon", "lat", "time"],
        name="data",
    ).transpose("time", "lon", "lat")
    return times_data


def generate_temperature_data():
    da = xr.Dataset()
    da["air"] = generate_data()
    return da


def generate_temperature_data_multi_time_scale():
    da = xr.Dataset()
    da["air"] = generate_data(n_time=12, freq="d").rename({"time": "day"})
    da["ground"] = generate_data(freq="H")
    return da



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py,md
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
# Uncomment to run the notebook in Colab
# # ! pip install -q "wax-ml[complete]@git+https://github.com/eserie/wax-ml.git"
# # ! pip install -q --upgrade jax jaxlib==0.1.70+cuda111 -f https://storage.googleapis.com/jax-releases/jax_releases.html
# -

# check available devices
import jax

print("jax backend {}".format(jax.lib.xla_bridge.get_backend().platform))
jax.devices()

# # 〰 Compute exponential moving averages with xarray and pandas accessors 〰
#
# [![Open in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/eserie/wax-ml/blob/main/docs/notebooks/01_demo_EWMA.ipynb)

# WAX-ML implements pandas and xarray accessors to ease the usage of machine-learning algorithms with
# high-level data APIs :
# - pandas's `DataFrame` and `Series`
# - xarray's `Dataset` and `DataArray`.
#
# These accessors allow to easily execute any function using Haiku modules
# on these data containers.
#
# For instance, WAX-ML propose an implementation of the exponential moving average realized
# with this mechanism.
#
# Let's show how it works.

# ## Load accessors

# First you need to load accessors:

# + tags=[]
from wax.accessors import register_wax_accessors

register_wax_accessors()
# -

# ## EWMA on dataframes

# Let's look at a simple example: The exponential moving average (EWMA).
#
# Let's apply the EWMA algorithm to the [NCEP/NCAR 's Air temperature data](http://www.esrl.noaa.gov/psd/data/gridded/data.ncep.reanalysis.html).

# ### 🌡 Load temperature dataset 🌡

# + tags=[]
import xarray as xr

dataset = xr.tutorial.open_dataset("air_temperature")
# -

# Let's see what this dataset looks like:

# + tags=[]
dataset
# -

# To compute a EWMA on some variables of a dataset, we usually need to convert data
# in pandas
# [series](https://pandas.pydata.org/docs/reference/api/pandas.Series.html) or
# [dataframe](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html).
#
# So, let's convert the dataset into a dataframe to illustrate `accessors` on a dataframe:

dataframe = dataset.air.to_series().unstack(["lon", "lat"])

# ### EWMA with pandas

air_temp_ewma = dataframe.ewm(alpha=1.0 / 10.0).mean()
_ = air_temp_ewma.iloc[:, 0].plot()

# ### EWMA with WAX-ML

air_temp_ewma = dataframe.wax.ewm(alpha=1.0 / 10.0).mean()
_ = air_temp_ewma.iloc[:, 0].plot()

# On small data, WAX-ML's EWMA is slower than Pandas' because of the expensive data conversion steps.
# WAX-ML's accessors are interesting to use on large data loads
# (See our [three-steps_workflow](https://wax-ml.readthedocs.io/en/latest/notebooks/04_The_three_steps_workflow.html))
#
# ## Apply a custom function to a Dataset

# Now let's illustrate how WAX-ML accessors work on [xarray datasets](http://xarray.pydata.org/en/stable/generated/xarray.Dataset.html).

# +
from wax.modules import EWMA


def my_custom_function(dataset):
    return {
        "air_10": EWMA(1.0 / 10.0)(dataset["air"]),
        "air_100": EWMA(1.0 / 100.0)(dataset["air"]),
    }


dataset = xr.tutorial.open_dataset("air_temperature")
output, state = dataset.wax.stream().apply(
    my_custom_function, format_dims=dataset.air.dims
)

_ = output.isel(lat=0, lon=0).drop(["lat", "lon"]).to_pandas().plot(figsize=(12, 8))



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as onp
import pandas as pd
import pytest

from wax.external.eagerpy import convert_to_tensors
from wax.universal.eager_ewma import (
    EagerEWMA,
    dynamic_unroll_fori_loop_universal,
    dynamic_unroll_tf,
    dynamic_unroll_universal,
    static_unroll_universal,
)


def test_init_and_first_step_ema_float64():
    from jax.config import config

    config.update("jax_enable_x64", True)
    seed = 1701
    key = jax.random.PRNGKey(seed)
    x = jax.random.normal(shape=(3,), key=key, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(0.1), adjust=True)(x)

    params, state = model.init(key, x)
    ema, state = model.apply(params, state, key, x)
    assert ema.dtype == jnp.dtype("float64")


def test_init_and_first_step_ema():
    from jax.config import config

    config.update("jax_enable_x64", False)

    seed = 1701
    key = jax.random.PRNGKey(seed)
    x = jax.random.normal(shape=(3,), key=key, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(0.1), adjust=True)(x)

    params, state = model.init(key, x)
    ema, state1 = model.apply(params, state, key, x)
    assert ema.dtype == jnp.dtype("float32")


def test_run_ema_vs_pandas_not_adjust():
    from jax.config import config

    config.update("jax_enable_x64", True)
    import pandas as pd

    seed = 1701
    key = jax.random.PRNGKey(seed)
    x = jax.random.normal(shape=(100, 3), key=key, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(0.1), adjust=False)(x)

    ema, state = static_unroll_universal(model, None, None, key, False, x)

    pandas_ema = pd.DataFrame(x).ewm(alpha=jnp.array(0.1), adjust=False).mean()

    assert jnp.allclose(ema, pandas_ema.values)


def test_dynamic_unroll_fori_loop():
    from jax.config import config

    config.update("jax_enable_x64", True)

    seed = 1701
    key = jax.random.PRNGKey(seed)
    x = jax.random.normal(shape=(100, 3), key=key, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(0.1), adjust=True)(x)

    ema, state = static_unroll_universal(model, None, None, key, False, x)

    ema2, state2 = dynamic_unroll_fori_loop_universal(model, None, None, key, False, x)

    assert jnp.allclose(ema, ema2)


def test_dynamic_unroll():
    from jax.config import config

    config.update("jax_enable_x64", True)

    seed = 1701
    key = jax.random.PRNGKey(seed)
    x = jax.random.normal(shape=(100, 3), key=key, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(0.1), adjust=True)(x)

    ema, state = static_unroll_universal(model, None, None, key, False, x)

    ema2, state2 = dynamic_unroll_universal(model, None, None, key, False, x)

    assert jnp.allclose(ema, ema2)


def test_run_ema_vs_pandas_adjust():
    from jax.config import config

    config.update("jax_enable_x64", True)

    seed = 1701
    key = jax.random.PRNGKey(seed)
    x = jax.random.normal(shape=(100, 3), key=key, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(0.1), adjust=True)(x)

    ema, state = dynamic_unroll_universal(model, None, None, key, False, x)

    pandas_ema = pd.DataFrame(x).ewm(alpha=jnp.array(0.1), adjust=True).mean()
    assert jnp.allclose(ema, pandas_ema.values)


def test_run_ema_vs_pandas_adjust_finite():
    from jax.config import config

    config.update("jax_enable_x64", True)

    seed = 1701
    key = jax.random.PRNGKey(seed)
    x = jax.random.normal(shape=(100, 3), key=key, dtype=jnp.float64)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(0.1), adjust="linear")(x)

    ema, state = dynamic_unroll_universal(model, None, None, key, False, x)

    pandas_ema_adjust = pd.DataFrame(x).ewm(alpha=jnp.array(0.1), adjust=True).mean()
    pandas_ema_not_adjust = (
        pd.DataFrame(x).ewm(alpha=jnp.array(0.1), adjust=True).mean()
    )
    assert not jnp.allclose(ema, pandas_ema_adjust.values)
    assert not jnp.allclose(ema, pandas_ema_not_adjust.values)
    corr = jnp.corrcoef(ema.flatten(), pandas_ema_adjust.values.flatten())[0, 1]
    assert 1.0e-3 < 1 - corr < 1.0e-2


@pytest.mark.parametrize("tensor_type", ["numpy", "jax", "tensorflow", "torch"])
def test_backends(tensor_type):
    from jax.config import config

    config.update("jax_enable_x64", True)

    key = jax.random.PRNGKey(403)
    ox = onp.random.normal(size=(100, 3))
    oalpha = onp.array(0.1)
    x, alpha = convert_to_tensors((ox, oalpha), tensor_type)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(alpha, adjust=False)(x)

    ema, state = static_unroll_universal(model, None, None, key, False, x)

    pandas_ema = pd.DataFrame(ox).ewm(oalpha, adjust=False).mean()
    df = pd.concat(
        [pd.DataFrame(ox), pd.DataFrame(ema), pandas_ema],
        keys=["input", tensor_type, "pandas"],
        axis=1,
    ).stack()
    df = df.xs(0, level=1)
    assert len(df)
    assert not onp.allclose(ema, pandas_ema.values)


def test_tf_optimized():
    tensor_type = "tensorflow"

    key = None  # jax.random.PRNGKey(403)
    ox = onp.random.normal(size=(100, 3))
    oalpha = 0.1
    x, alpha = convert_to_tensors((ox, oalpha), tensor_type)

    @hk.transform_with_state
    def model(x):
        return EagerEWMA(jnp.array(alpha), adjust=False)(x)

    ema, state = dynamic_unroll_tf(model, None, None, key, False, x)

    pandas_ema = pd.DataFrame(ox).ewm(oalpha, adjust=False).mean()
    df = pd.concat(
        [pd.DataFrame(ox), pd.DataFrame(ema), pandas_ema],
        keys=["input", tensor_type, "pandas"],
        axis=1,
    ).stack()
    df = df.xs(0, level=1)
    assert len(df)
    assert not onp.allclose(ema, pandas_ema.values)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from tqdm.auto import tqdm

from wax.gym.callbacks.callbacks import Callback


class ProgressBar(Callback):
    def __init__(self):
        # state
        self._pbar = None

    def __enter__(self):
        super().__enter__()

    def _on_train_start(self, env, agent, obs):
        if self._pbar is None:
            try:
                self._pbar = tqdm(total=len(env))
            except TypeError:
                self._pbar = tqdm()

    def _on_step(self, env, agent, gym_state):
        self._pbar.update()

    def _on_train_end(self, env, agent):
        if self._pbar is not None:
            self._pbar.close()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import pytest

from wax.modules.mask_std import MaskStd


def test_mask_std_no_mask():

    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (10,))
    std_ref = x.std()

    fun = hk.transform(lambda mask, x: MaskStd()(mask, x))
    mask = jnp.full(x.shape, True)

    params = fun.init(next(rng), mask, x)
    x_std = fun.apply(params, next(rng), mask, x)
    assert jnp.allclose(x_std, std_ref)


@pytest.mark.parametrize("assume_centered", [False, True])
def test_mask_std_with_mask(assume_centered):

    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (10,))
    if assume_centered:
        x_std_ref = jnp.sqrt((x[1:] ** 2).mean())
    else:
        x_std_ref = x[1:].std()

    fun = hk.transform(
        lambda mask, x: MaskStd(assume_centered=assume_centered)(mask, x)
    )

    mask = jnp.full(x.shape, True)
    mask = jax.ops.index_update(mask, jax.ops.index[:1], False)

    params = fun.init(next(rng), mask, x)
    x_std = fun.apply(params, next(rng), mask, x)
    assert jnp.allclose(x_std, x_std_ref)


def test_mask_std_with_mask_shape_2():
    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (2, 3))

    fun = hk.transform(lambda mask, x: MaskStd()(mask, x))

    mask = jnp.full(x.shape, True)
    mask = jax.ops.index_update(mask, jax.ops.index[0, 2], False)

    params = fun.init(next(rng), mask, x)
    x_std = fun.apply(params, next(rng), mask, x)

    # std_ref = jax.ops.index_update(x, ~mask, 0).std()
    x_std_ref = x[mask].std()
    assert jnp.allclose(x_std, x_std_ref)


def test_mask_std_axis_1():
    rng = hk.PRNGSequence(42)
    x = jax.random.normal(next(rng), (2, 3))

    fun = hk.transform(lambda mask, x: MaskStd(axis=0)(mask, x))

    mask = jnp.full(x.shape, True)
    mask = jax.ops.index_update(mask, jax.ops.index[0, 2], False)

    params = fun.init(next(rng), mask, x)
    x_std = fun.apply(params, next(rng), mask, x)

    # the last axis has only one observation
    assert jnp.allclose(x_std[2], 0.0)

    for col in range(3):
        x_std_col_ref = x[:, col][mask[:, col]].std()
        assert jnp.allclose(x_std[col], x_std_col_ref)


def test_grad_mask_std():
    rng = jax.random.PRNGKey(42)

    shape = (10, 2)
    x = jax.random.normal(rng, shape)

    _, rng = jax.random.split(rng)
    mask = jax.random.choice(rng, 2, shape=shape).astype(bool)

    _, rng = jax.random.split(rng)
    params = {"w": jax.random.normal(rng, shape)}

    # put some nan values
    x = jax.ops.index_update(x, 0, jnp.nan)

    @hk.transform
    def fun(x):
        w = hk.get_parameter("w", x.shape, init=lambda *_: jnp.zeros_like(x))
        y = w * x
        score = MaskStd()(mask, y)
        return score

    params = fun.init(rng, x)
    scrore, grads = jax.value_and_grad(fun.apply)(params, rng, x)
    assert not jnp.isnan(grads["~"]["w"]).all()


def test_grad_mask_std_unroll():
    rng = jax.random.PRNGKey(42)

    shape = (10, 2)
    x = jax.random.normal(rng, shape)

    _, rng = jax.random.split(rng)
    mask = jax.random.choice(rng, 2, shape=shape).astype(bool)

    _, rng = jax.random.split(rng)
    params = {"w": jax.random.normal(rng, shape)}

    # put some nan values
    x = jax.ops.index_update(x, 0, jnp.nan)

    from wax.unroll import unroll_transform_with_state

    @unroll_transform_with_state
    def fun(mask, x):
        w = hk.get_parameter("w", x.shape, init=lambda *_: jnp.zeros_like(x))
        return MaskStd()(mask, (w * jnp.nan_to_num(x)))

    def batch(params, state, rng, mask, x):
        score, state = fun.apply(params, state, rng, mask, x)
        return jnp.nanmean(score)

    params, state = fun.init(rng, mask, x)
    scrore, grads = jax.value_and_grad(batch)(params, state, rng, mask, x)
    assert not jnp.isnan(grads["~"]["w"]).all()



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from jax import tree_flatten, tree_unflatten

from .tensor import (
    JAXTensor,
    NumPyTensor,
    PyTorchTensor,
    Tensor,
    TensorFlowTensor,
    TensorType,
)
from .types import NativeTensor

if TYPE_CHECKING:
    # for static analyzers
    import torch


def _get_module_name(x: Any) -> str:
    # splitting is necessary for TensorFlow tensors
    return type(x).__module__.split(".")[0]


@overload
def astensor(x: TensorType) -> TensorType:
    ...


@overload
def astensor(x: "torch.Tensor") -> PyTorchTensor:
    ...


@overload
def astensor(x: NativeTensor) -> Tensor:  # type: ignore
    ...


def astensor(x: Union[NativeTensor, Tensor, Any]) -> Union[Tensor, Any]:  # type: ignore
    # we use the module name instead of isinstance
    # to avoid importing all the frameworks
    name = _get_module_name(x)

    if name == "torch":
        return PyTorchTensor(x)  # type: ignore
    if name == "tensorflow":
        return TensorFlowTensor(x)
    if name == "jax" or name == "jaxlib":
        return JAXTensor(x)  # type: ignore
    if name == "numpy":
        return NumPyTensor(x)  # type: ignore
    if isinstance(x, (str, Number)):
        return NumPyTensor(x)  # type: ignore

    # non Tensor types are returned unmodified
    return x


def astensors(data: Any, *args) -> Any:  # type: ignore
    if args:
        data = (data,) + args
    leaf_values, tree_def = tree_flatten(data)
    leaf_values = tuple(astensor(value) for value in leaf_values)
    return tree_unflatten(tree_def, leaf_values)


T = TypeVar("T")


class RestoreTypeFunc(Generic[T]):
    def __init__(self, x: T):
        self.unwrap = not isinstance(x, Tensor)

    @overload
    def __call__(self, x: Tensor) -> T:
        ...

    @overload  # noqa: F811
    def __call__(self, x: Tensor, y: Tensor) -> Tuple[T, T]:
        ...

    @overload  # noqa: F811
    def __call__(self, x: Tensor, y: Tensor, z: Tensor, *args: Tensor) -> Tuple[T, ...]:
        ...

    @overload  # noqa: F811
    def __call__(self, *args: Any) -> Any:
        # catch other types, otherwise we would return type T for input type Any
        ...

    def __call__(self, *tensors):  # type: ignore  # noqa: F811
        result = tuple(self._restore(x) for x in tensors) if self.unwrap else tensors
        if len(result) == 1:
            (result,) = result
        return result

    def _restore(self, x):  # type: ignore  # noqa: F811
        x_raw = as_raw_tensors(x)
        if isinstance(x_raw, tuple):
            assert len(x_raw) == 2
            x_raw, unwrap = x_raw
            assert unwrap
            return x_raw
        else:
            return x_raw


def astensor_(x: T) -> Tuple[Tensor, RestoreTypeFunc[T]]:
    return astensor(x), RestoreTypeFunc[T](x)


def astensors_(x: T, *xs: T) -> Tuple[Tuple[Tensor, ...], RestoreTypeFunc[T]]:
    return astensors(x, *xs), RestoreTypeFunc[T](x)


def has_tensor(tree_def: Any) -> bool:
    if "NumPyTensor" in str(tree_def):
        return True
    if "JAXTensor" in str(tree_def):
        return True
    if "TensorFlowTensor" in str(tree_def):
        return True
    if "PyTorchTensor" in str(tree_def):
        return True
    return False


def as_tensors_any(data: Any) -> Tuple[Any, bool]:
    """Convert data structure leaves in Tensor and detect if any of the input data contains a Tensor.

    Parameters
    ----------
    data
        data structure.

    Returns
    -------
    Any
        modified data structure.
    bool
        True if input data contains a Tensor type.
    """
    leaf_values, tree_def = tree_flatten(data)
    transformed_leaf_values = tuple(astensor(value) for value in leaf_values)
    return tree_unflatten(tree_def, transformed_leaf_values), has_tensor(tree_def)


def _is_tensor(x: T) -> bool:
    name = _get_module_name(x)

    if name == "torch":
        return True
    if name == "tensorflow":
        return True
    if name == "jax" or name == "jaxlib":
        return True
    if name == "numpy":
        return True
    if isinstance(x, (str, int)):
        return True
    return False


def as_raw_tensor_leave(x: T) -> Any:
    if _is_tensor(x):
        unwrap = True
        return (x, unwrap)
    else:
        return x


def as_raw_tensor(x: T) -> Any:
    if isinstance(x, Tensor):
        unwrap = True
        return (x.raw, unwrap)
    else:
        return x


def as_raw_tensors(data: Any) -> Any:
    leaf_values, tree_def = tree_flatten(data)
    if not has_tensor(tree_def):
        return data
    leaf_values = tuple(as_raw_tensor(value) for value in leaf_values)
    leaf_values_raw = list(map(as_raw_tensor_leave, leaf_values))
    data_raw = tree_unflatten(tree_def, leaf_values_raw)
    return data_raw


def eager_function(func: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(func)
    def eager_func(*args: Any, **kwargs: Any) -> Any:
        (args, kwargs), has_tensor = as_tensors_any((args, kwargs))
        unwrap = not has_tensor
        result = func(*args, **kwargs)
        if unwrap:
            raw_result = as_raw_tensors(result)
            return raw_result
        else:
            return result

    return eager_func



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as onp
import pytest

from wax.compile import jit_init_apply
from wax.modules.ewma import EWMA
from wax.modules.ewmvar import EWMVar
from wax.unroll import unroll


# Another implementation for checking
class EWMVar_v2(hk.Module):
    """Exponentially weighted variance.
    To calculate the variance we use the fact that Var(X) = Mean(x^2) - Mean(x)^2 and internally
    we use the exponentially weighted mean of x/x^2 to calculate this.

    Arguments:
        alpha : The closer `alpha` is to 1 the more the statistic will adapt to recent values.

    Attributes:
        variance : The running exponentially weighted variance.

    References
    ----------
    [^1]: [Finch, T., 2009. Incremental calculation of weighted mean and variance. University of Cambridge, 4(11-5), pp.41-42.](https://fanf2.user.srcf.net/hermes/doc/antiforgery/stats.pdf) # noqa
    """

    def __init__(self, alpha=0.5, adjust=True, name=None):
        super().__init__(name=name)
        self.alpha = alpha
        self.adjust = adjust

    def __call__(self, x):
        mean = EWMA(self.alpha, self.adjust, initial_value=jnp.nan, name="mean")(x)
        mean_square = EWMA(
            self.alpha, self.adjust, initial_value=jnp.nan, name="mean_square"
        )(x * x)
        var = mean_square - mean ** 2
        var = jnp.where(var < 0, 0.0, var)
        return var


@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_init_and_first_step_var_float64(dtype):
    from jax.config import config

    if dtype == "float64":
        config.update("jax_enable_x64", True)
    else:
        config.update("jax_enable_x64", False)
    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(3,), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMVar(0.1, adjust=True)(x)

    params, state = model.init(next(seq), x)
    var, state = model.apply(params, state, next(seq), x)
    assert var.dtype == jnp.dtype(dtype)


def test_run_var_vs_pandas_not_adjust():
    from jax.config import config

    config.update("jax_enable_x64", True)
    import pandas as pd

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMVar(0.1, adjust=False)(x)

    var, state = unroll(model, return_final_state=True)(x)
    var = pd.DataFrame(var)

    @jit_init_apply
    @hk.transform_with_state
    def model2(x):
        return EWMVar_v2(0.1, adjust=False)(x)

    var2, state2 = unroll(model2, return_final_state=True)(x)
    var2 = pd.DataFrame(var2)
    assert onp.allclose(var, var2)

    pandas_var = pd.DataFrame(x).ewm(alpha=0.1, adjust=False).var()
    assert not onp.allclose(var, pandas_var.values)


def test_run_var_vs_pandas_adjust():
    from jax.config import config

    config.update("jax_enable_x64", True)
    import pandas as pd

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMVar(0.1, adjust=True)(x)

    var, state = unroll(model, return_final_state=True)(x)
    var = pd.DataFrame(var)

    @jit_init_apply
    @hk.transform_with_state
    def model2(x):
        return EWMVar_v2(0.1, adjust=True)(x)

    var2, state2 = unroll(model2, return_final_state=True)(x)
    var2 = pd.DataFrame(var2)
    assert onp.allclose(var, var2)

    # pandas does something else
    pandas_var = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).var()
    assert not onp.allclose(var, pandas_var.values)


def test_run_var_vs_pandas_adjust_finite():
    from jax.config import config

    config.update("jax_enable_x64", True)
    import pandas as pd

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x):
        return EWMVar(0.1, adjust="linear")(x)

    var, state = unroll(model, return_final_state=True)(x)
    var = pd.DataFrame(var)

    @jit_init_apply
    @hk.transform_with_state
    def model2(x):
        return EWMVar_v2(0.1, adjust=True)(x)

    var2, state2 = unroll(model2, return_final_state=True)(x)
    var2 = pd.DataFrame(var2)
    assert not onp.allclose(var, var2)
    # TODO: understand why the two implementations are
    #  not agree for "linear" adjustement scheme.

    pandas_var_adjust = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).var()
    pandas_var_not_adjust = pd.DataFrame(x).ewm(alpha=0.1, adjust=True).var()
    assert not onp.allclose(var, pandas_var_adjust.values)
    assert not onp.allclose(var, pandas_var_not_adjust.values)
    corr = jnp.corrcoef(
        var.fillna(0).values.flatten(), pandas_var_adjust.fillna(0).values.flatten()
    )[0, 1]
    assert 0.08 < 1 - corr < 0.1



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Union

from .framework import inf
from .tensor import TensorType
from .types import AxisAxes


def l0(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return (x != 0).sum(axis=axis, keepdims=keepdims).astype(x.dtype)


def l1(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return x.abs().sum(axis=axis, keepdims=keepdims)


def l2(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return x.square().sum(axis=axis, keepdims=keepdims).sqrt()


def linf(
    x: TensorType, axis: Optional[AxisAxes] = None, keepdims: bool = False
) -> TensorType:
    return x.abs().max(axis=axis, keepdims=keepdims)


def lp(
    x: TensorType,
    p: Union[int, float],
    axis: Optional[AxisAxes] = None,
    keepdims: bool = False,
) -> TensorType:
    if p == 0:
        return l0(x, axis=axis, keepdims=keepdims)
    if p == 1:
        return l1(x, axis=axis, keepdims=keepdims)
    if p == 2:
        return l2(x, axis=axis, keepdims=keepdims)
    if p == inf:
        return linf(x, axis=axis, keepdims=keepdims)
    return x.abs().pow(p).sum(axis=axis, keepdims=keepdims).pow(1.0 / p)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp
import pytest
from jax.config import config
from sklearn.covariance import EmpiricalCovariance

from wax.compile import jit_init_apply
from wax.modules.ewmcov import EWMCov
from wax.unroll import unroll


@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_init_and_first_step_cov_float64(dtype):
    if dtype == "float64":
        config.update("jax_enable_x64", True)
    else:
        config.update("jax_enable_x64", False)
    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(3,), key=next(seq), dtype=jnp.float64)
    y = jax.random.normal(shape=(3,), key=next(seq), dtype=jnp.float64)

    @jit_init_apply
    @hk.transform_with_state
    def model(x, y):
        return EWMCov(jnp.array(0.1), adjust=True)(x, y)

    params, state = model.init(next(seq), x, y)
    cov, state = model.apply(params, state, next(seq), x, y)
    assert cov.dtype == jnp.dtype(dtype)


@pytest.mark.parametrize("assume_centered", [False, True])
def test_run_cov_vs_sklearn(assume_centered):
    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)
    y = x

    alpha = 1.0e-6

    @jit_init_apply
    @hk.transform_with_state
    def model(x, y):
        return EWMCov(alpha, adjust=True, assume_centered=assume_centered)(x, y)

    cov = unroll(model)(x, y)
    cov_ref = EmpiricalCovariance(assume_centered=assume_centered).fit(x).covariance_

    assert jnp.allclose(cov[-1], cov_ref, atol=1.0e-6)


@pytest.mark.parametrize("assume_centered", [False, True])
def test_run_cov_vs_sklearn_adjust(assume_centered):
    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)
    y = x

    alpha = 1.0e-6

    @jit_init_apply
    @hk.transform_with_state
    def model(x, y):
        return EWMCov(alpha, adjust=True, assume_centered=assume_centered)(x, y)

    cov = unroll(model)(x, y)
    cov_ref = EmpiricalCovariance(assume_centered=assume_centered).fit(x).covariance_

    assert jnp.allclose(cov[-1], cov_ref, atol=1.0e-6)


@pytest.mark.parametrize("assume_centered", [False, True])
def test_run_cov_vs_pandas_adjust_finite(assume_centered):
    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)
    y = x

    alpha = 1.0e-6

    @jit_init_apply
    @hk.transform_with_state
    def model(x, y):
        return EWMCov(alpha, adjust="linear", assume_centered=assume_centered)(x, y)

    cov = unroll(model)(x, y)
    cov_ref = EmpiricalCovariance(assume_centered=assume_centered).fit(x).covariance_

    assert jnp.allclose(cov[-1], cov_ref, atol=1.0e-6)


@pytest.mark.parametrize("assume_centered", [False, True])
def test_run_cov_with_legacy_api(assume_centered):
    config.update("jax_enable_x64", True)

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(10, 3), key=next(seq), dtype=jnp.float64)
    y = x

    alpha = 1 / 10.0

    @jit_init_apply
    @hk.transform_with_state
    def model(x, y):
        return EWMCov(alpha, adjust="linear", assume_centered=assume_centered)(x, y)

    cov = unroll(model)(x, y)

    @jit_init_apply
    @hk.transform_with_state
    def model(x_y):
        return EWMCov(alpha, adjust="linear", assume_centered=assume_centered)(x_y)

    cov_ref = unroll(model)(x_y=(x, y))

    assert jnp.allclose(cov, cov_ref, atol=1.0e-6)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
from typing import Any

from jax.tree_util import tree_flatten, tree_unflatten

from .astensor import _get_module_name


def convert_to_tensor(data: Any, tensor_type: str) -> Any:
    """Convert a tensor in a given tensor_type.

    Parameters
    ----------
    tensor_type
        The targeted tensor type. Can be in ['numpy', 'tensorflow', 'jax', 'torch"].

    Returns
    -------
    data
        data structure with converted tensors.
    """
    name = _get_module_name(data)
    if name not in ["numpy", "jax", "torch", "tensorflow"]:
        # do not convert
        return data

    if tensor_type == "tensorflow":
        import tensorflow as tf

        return tf.convert_to_tensor(data)

    elif tensor_type == "torch":
        import torch

        return torch.tensor(data)

    elif tensor_type == "jax":
        import jax.numpy as jnp

        return jnp.asarray(data, dtype=data.dtype)

    elif tensor_type == "numpy":
        import numpy as onp

        return onp.asarray(data)

    raise ValueError(
        f"tensor_type {tensor_type} must be in ['numpy', 'tensorflow', 'jax', 'torch']"
    )


def convert_to_tensors(data: Any, tensor_type: str) -> Any:
    """Convert tensors in a nested data structure .

    Parameters
    ----------
    tensor_type
        The targeted tensor type. Can be in ['numpy', 'tensorflow', 'jax', 'torch"].

    Returns
    -------
    data
        data structure with converted tensors.
    """
    leaf_values, tree_def = tree_flatten(data)
    leaf_values = list(
        map(partial(convert_to_tensor, tensor_type=tensor_type), leaf_values)
    )
    return tree_unflatten(tree_def, leaf_values)



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk
import jax
import jax.numpy as jnp

from wax.modules.fill_nan_inf import FillNanInf


def test_fill_nan_inf():

    seq = hk.PRNGSequence(42)
    x = jax.random.normal(shape=(100, 5), key=next(seq), dtype=jnp.float64)
    mask_nan = jax.random.choice(next(seq), 2, shape=x.shape)
    x = jnp.where(mask_nan, x, jnp.nan)
    del mask_nan

    mask_inf = jax.random.choice(next(seq), 2, shape=x.shape)
    x = jnp.where(mask_inf, x, jnp.inf)
    del mask_inf

    mask_ninf = jax.random.choice(next(seq), 2, shape=x.shape)
    x = jnp.where(mask_ninf, x, -jnp.inf)
    del mask_ninf

    fun = hk.transform(lambda x: FillNanInf()(x))

    params = fun.init(next(seq), x)
    x_fill = fun.apply(params, next(seq), x)
    assert jnp.allclose(x_fill, jnp.nan_to_num(x, nan=0.0, posinf=0.0, neginf=0.0))



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import haiku as hk


class UpdateParams(hk.Module):
    def __init__(self, fun, name=None):
        super().__init__(name=name)
        self.fun = (
            fun
            if isinstance(fun, hk.TransformedWithState)
            else hk.transform_with_state(fun)
        )

    def __call__(self, params, *args, **kwargs):
        rng = hk.next_rng_key()

        def init_state(*_):
            _, state = self.fun.init(rng, *args, **kwargs)
            return state

        state = hk.get_state("state", [], init=init_state)
        res, state = self.fun.apply(params, state, rng, *args, **kwargs)
        hk.set_state("state", state)

        return res



# Copyright 2021 The WAX-ML Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Rolling mean."""
import haiku as hk
import jax.numpy as jnp

from wax.modules.buffer import Buffer


class RollingMean(hk.Module):
    """Rolling mean."""

    def __init__(self, horizon: int, min_periods: int = 1, name: str = None):
        """Initialize the module.

        Args:
            horizon: horizon on which we compute the mean.
            min_periods: minimum number of data point required in the window to caompute the mean.
            name: name of the module instance.
        """
        super().__init__(name=name)
        self.horizon = horizon
        self.min_periods = min_periods

    def __call__(self, x):
        buffer, attrs = Buffer(self.horizon, return_state=True)(x)
        sum = jnp.where(jnp.logical_not(jnp.isnan(buffer)), buffer, 0.0).sum(axis=0)
        count = jnp.where(jnp.logical_not(jnp.isnan(buffer)), 1, 0).sum(axis=0)
        mean = jnp.where(count >= self.min_periods, sum / count, jnp.nan)
        return mean
